module(..., package.seeall)
local s = require'scheme'
local nsg = require'nsg'
local iface = require'nsgifaceconf'
local ts = require'nsgtsconf'
local lpeg = require'lpeg'
local conflib = require'nsgconflib'
nsg.meta()


local stateList = {}

function registerVirtSensor(sensor, states)
  for i = 1, #states do
    states[i] = states[i]:match"^(.*)&&" or states[i]
  end
  stateList[sensor] = states
end

function unregisterVirtSensor(sensor)
  stateList[sensor] = nil
end

-- event handler for netlink
local ifStateEventTable = {}
local function ifStateEvent(msg)
  local msgtype = msg.type
  if msgtype == 'newlink' or msgtype == 'dellink' then
    local ifname, ifup = msg.ifname, msg.ifup
    local ifState = msgtype == 'dellink' and 'no' or (ifup and 'up' or 'down')
    if ifStateEventTable[ifname] ~= ifState then
      ifStateEventTable[ifname] = ifState
      local ifSensor = 'ifstate.'..ifname
      eventHandler(ifSensor, ifState)
      local alias = iface.getAliasName(ifname)
      if alias and alias ~= ifname then
        eventHandler('ifstate.'..alias, ifState)
      end
    end
  end
end

registerVirtSensor('ifstate.', {'up', 'down', 'no'})

-- scheduler
local getSet do
  local MIN, MAX = 0, 59
  local ERR
  local function err(s, i)
    ERR = ("error near %q"):format(s:sub(i, 20))
  end

  local function makeNumberSet(n)
    local t = {}
    if n >= MIN and n <= MAX then
      t[n] = true
    end
    return t
  end

  local function makeRangeSet(a, b, s)
    s = s or 1
    b = b <= MAX and b or MAX
    local t = {}
    for i = a, b, s do
      if i >= MIN then
        t[i] = true
      end
    end
    return t
  end

  local function makeAnySet(_, s)
    return makeRangeSet(0, 59, s)
  end

  local function merge(t1, t2)
    if t2 then
      for k, _ in pairs(t2) do
        t1[k] = true
      end
    end
    return t1
  end

  local sp = lpeg.S" \n\r\t"^0
  local Number = lpeg.R("09")^1 / tonumber
  local Step = '/' * sp * Number
  local NumberSet = Number / makeNumberSet
  local Range =  Number * sp * ".." * sp * Number * (sp * Step)^-1
  local RangeSet = Range / makeRangeSet
  local Any = lpeg.C"*" * (sp * Step)^-1
  local AnySet = Any / makeAnySet
  local Element = sp * (RangeSet + AnySet + NumberSet + err)
  local Set = lpeg.Cf(Element * (sp * ',' * Element)^0 * sp * (lpeg.P(-1) + err), merge)

  function getSet(str, min, max)
    MIN = min or 0
    MAX = max or 59
    ERR = nil
    return Set:match(str), ERR
  end
end


local function cronValidate(val)
  val = tostring(val)
  local res, err = getSet(val)
  if res then
    return val
  else
    return nil, "string(see manual)\n"..err
  end
end

local scriptPattern = [[
local scheduler = %s
local socket = require'socket'
while true do
  local t = os.date'*t'
  for  _, v in ipairs(scheduler) do
    if (
      v[1][t.min] and v[2][t.hour] and v[3][t.day] and v[4][t.month] and v[5][t.wday]
    ) then
      print('scheduler', v.state)
    end
  end
  socket.sleep(60 - t.sec)
end
]]

local function getSchedTable(conf)
  local res = {}
  local states = {}
  for _, v in ipairs(conf) do 
    if v.enable then
      local t = {}
      t.state = v['event-name']
      t[1] = getSet(v.min, 0, 59)
      t[2] = getSet(v.hour, 0, 23)
      t[3] = getSet(v.day, 1, 31)
      t[4] = getSet(v.month, 1, 12)
      t[5] = getSet(v.wday, 1, 7)
      res[#res+1] = t
      states[#states+1] = t.state
    end
  end
  return nsg.serialize(res, 0), states
end

local function schedulerApply(path, sch, conf, prevConf)
  if not s.confCmp(conf, prevConf) then
    local conf = s.addDefault(path, sch, conf, true) --clone conf with defaults

    local enable = false
    for _, v in ipairs(conf) do
      if v.enable then
	enable = true
	break
      end
    end
    if not enable then
      C.restartDaemon(path)
      unregisterVirtSensor('scheduler')
      return true
    end

    local stab, states = getSchedTable(conf)
    local script = scriptPattern:format(stab)

    local com = {
      "lua", "-e", script,
    }
    restartVirtSensor(path, com)
    registerVirtSensor('scheduler', states)
  end
  return true
end

---- timer
local timerScriptPattern = [[
local socket = require"socket"
local state, tmo = %q, %d
for i=1,%s do
  socket.sleep(tmo)
  print("timer", state)
end
]]

local aliveTimers = {}

local function timerApply(path, sch, rawConf, prevConf)
  if not s.confCmp(rawConf, prevConf) then
    local conf = s.addDefault(path, sch, rawConf, true) --clone conf with defaults
    local states = {}
    local timers = {}
    prevConf = prevConf or {}

    for k, cnf in pairs(conf) do
      timers[k] = true
      if not s.confCmp(rawConf[k], prevConf[k]) then
	local pth = s.clonePath(path, nil, k)
	if not cnf.enable then
	  C.restartDaemon(pth)
	else
	  local state = cnf["event-name"] or "tick"
	  local num = type(cnf.enable)=='number' and tostring(cnf.enable)
	  local script = timerScriptPattern:format(state, cnf.timeout or 10,
	                                           num or "math.huge")
	  states[state] = true

	  local com = {
	    "lua", "-e", script,
	  }
	  restartVirtSensor(pth, com, num)
	end
      end
    end

    for k in pairs(aliveTimers) do
      if not timers[k] then
	C.restartDaemon(s.clonePath(path, nil, k))
      end
    end
    aliveTimers = timers

    local stateList = {}
    for k in pairs(states) do
      stateList[#stateList+1] = k
    end
    if #stateList > 0 then
      registerVirtSensor('timer', stateList)
    else
      unregisterVirtSensor('timer')
    end
  end
  return true
end

-- action parser
local parseActionField do
  local ERR
  local function err(s, i)
    ERR = ("internal script syntax error near %q"):format(s:sub(i, 20))
  end

  local P  = lpeg.P
  local Ca = lpeg.C
  local Cs = lpeg.Cs
  local Ct = lpeg.Ct
  local S  = lpeg.S
  local V  = lpeg.V

  local sp = S" \n\r\t;"
  local eq = P"="
  local qt_char = '"'
  local qt = P(qt_char)
  local lp = P'('
  local rp = P')'
  local sep = lp + rp + sp + eq

  local unquoted_field = Ca( (1-sep) ^1 )
  local quoted_field   = qt * Cs( ((P(1)-qt) + (qt*qt)/qt_char) ^0) * qt
  local field          = quoted_field + unquoted_field
  local assign         = (field * sp^0 * eq * sp^0 * field)

  local prog = P{
    "prog",
    prog = Ct(sp^0 * V'scr'  * (sp^1 * V'scr')^0 * sp^0),
    rscr = field * sp^0 * lp * V'prog' * rp,
    scr = Ct(V'rscr' + assign + field + err),
  } * (P(-1) + err)

  local script
  local function parse(tbl, prefix)
    local firstT = type(tbl[1])
    if firstT == 'string' then
      prefix = prefix and prefix..'.'..tbl[1] or tbl[1]
      if #tbl == 1 then
        script[#script+1] = {prefix}
      elseif #tbl == 2 then
        secondT = type(tbl[2])
        if secondT == 'string' then
          script[#script+1] = {prefix, tbl[2]}
        elseif secondT == 'table' then
          parse(tbl[2], prefix)
        else
          return nil, nsg.serialize(tbl)
        end
      else
        return nil, nsg.serialize(tbl)
      end
    elseif firstT == 'table' then
      for _, v in ipairs(tbl) do
        parse(v, prefix)
      end
    else
      return nil, nsg.serialize(tbl)
    end
    return script
  end

  function parseActionField(act)
    act = act or ""
    local tbl = lpeg.match(prog, act)
    if tbl then
      script = {}
      local res, err = parse(tbl)
      return res, err
    else
      --local err = "Can't parse command: %s" % act
      C.log_error(ERR)
      return nil, ERR
    end
  end
end

--
local eventSesNum
local eventHandlerTable = {}
local sensorState = {}

local function openInternalSession()
  if not eventSesNum then
    local err
    eventSesNum, err = C.createNewSession{
      user = 'root', 
      maxRights = 'admin', 
      script = true
    }
    if not eventSesNum then
      C.log_error("Can't create internal session(%s)"%err)
    end
  end
  return eventSesNum
end

local function execInternalCommand(pathStr, val)
  local path = C.createPath(pathStr)
  local rootSes = C.S[0].ROOT_SESSION or eventSesNum
  if not rootSes then
    openInternalSession()
    if not eventSesNum then 
      return nil, "Can't open internal session"
    end
    rootSes = eventSesNum
  end
  local res, err, arg, ttyPath
  if tostring(path[#path]):sub(1,2) == '_a' then
    path[#path] = nil
    res, err = C.S[rootSes]:applyConf(path)
  else
    res, err, arg, ttyPath = C.S[rootSes]:setValueSmart(path, val)
  end
  if res == false and err == C.SESSION_IS_CLOSED then
    eventSesNum = nil
    return execInternalCommand(scr)
  end

  return res, err, arg, ttyPath
end

local function execInternalScript(scr)
  local commands, err = parseActionField(scr)
  if not commands then

    return nil, err
  end

  local res = {}
  for _, com in ipairs(commands) do
    local _, err, a, tty = execInternalCommand(com[1], com[2])
    res[#res+1] = com[1]..' '..tostring(err)..(tty and ' START DAEMON(See system.log.show)'or'')
  end
  if eventSesNum then
    C.S[eventSesNum]:closeSession()
    eventSesNum = nil
  end
  return table.concat(res, '\n')
end

local eventAction = {
  tcpsender = function(par, sensor, prevstate, state)
    return ts.tcpsenderSend(par, table.concat({sensor, prevstate, state}, ','))
  end,
}

local function execEventAction(scr, sensor, prevstate, state)
  local act, par = scr:match"^#([^%.]*)%.(.*)$"
  if act and eventAction[act] then
    return eventAction[act](par, sensor, prevstate, state)
  end
  return "Bad event action"
end

local function execScript(scr, ...)
  if not scr then
    return nil, "No script"
  end
  local t = scr:sub(1, 1)
  if t == '.' then
    return execInternalScript(scr)
  elseif t == '#' then
    return execEventAction(scr, ...)
  elseif scr:find"&%s*$" then
    return os.execute(("sh -c %q &"):format(scr:match"^(.*)&%s*$"))
  else
    return s.readpopen(scr.." 2>&1")
  end
end

local eventHandlerLogAdd, eventHandlerLogShow do
  local logtable = {}
  local first = 1
  local last = 1
  eventHandlerLogAdd = function(str)
    logtable[last] = str
    last = last + 1
    while first < last - 64 do
      logtable[first] = nil
      first = first + 1
    end
  end
  eventHandlerLogShow = function()
    return table.concat(logtable, '\n', first, last - 1)
  end
end

function eventHandler(sensor, fullstate)
  local state = fullstate:match"^(.*)&&" or fullstate
  local hn = {}
  local ss =  sensorState[sensor]
  for k, v in ipairs(eventHandlerTable) do
    if (sensor:find(v.sensor)
      and (
        (
          v.state == state 
          and (
            v.prev == ss or v.prev == 'any' 
            or (v.prev == 'other' and state ~= ss)
          )
        )
        or (
          v.prev == ss
          and (
            v.state == 'any' 
            or (v.state == 'other' and state ~= ss)
          )
        )
        or (
          v.prev == 'any' and v.state == 'any'
        )
        or (
          (v.prev == 'other' or v.prev == 'any')
          and (v.state == 'other' or v.state == 'any')
          and state ~= ss
        )
      )
    ) then
      execScript(v.action, sensor, tostring(ss), fullstate)
      hn[#hn+1] = k
    end
  end

  local logstr = "%s: %s --> %s [%s]" % {
    sensor, tostring(ss), state, table.concat(hn, ', ')
  }
  C.log_debug("eventHandler %s", logstr)
  eventHandlerLogAdd("%s %s" % {os.date('%F %T'), logstr})
  sensorState[sensor] = state
  return logstr
end

function restartVirtSensor(path,com,norestart)
  local myLog = ""
  C.restartDaemon(path, com,
    {
      noTimestamp = true,
      noRestart = norestart,
      atRead = function(oper)
        myLog = myLog..oper.log:read(-1)
        repeat
          local sensor, event = myLog:match"^([^\t\n\r:]+)[\t:]([^\n\r]+)[\n\r]"
          if sensor then
            eventHandler(sensor, event)
          end
          myLog = myLog:gsub("^[^\n\r]*[\n\r]+", "", 1)
        until not myLog:find"[\n\r]"
      end,
      atExit = C.increasingSleep,
    }
  ) 
end

local aliveEventGenerators = {}
local function restartEventGenerator(path, sch, conf, prevConf)
  if not s.confCmp(conf, prevConf) then
    local conf = s.addDefault(path, sch, conf, true)
    C.restartDaemon(path)

    if conf["enable"] then
      restartVirtSensor(path, {"/bin/sh", "-c", conf.command})
      aliveEventGenerators[path[#path]] = true
    end
  end
  return true
end

local function restartEventGenerators(path, sch, conf, prevConf)
  for k in pairs(aliveEventGenerators) do
    if not conf[k] then
      C.restartDaemon(s.clonePath(path, nil, k))
    end
  end
  return true
end

local function getStatelist(sensor)
  return sensor and (stateList[sensor] or 
    stateList[sensor:match"^(.*%.)[^%.]*$" or ''])
end

local stateScheme = function(h, p)
  return function(_, path)
    local sensor = C.getConfOrDefault(s.clonePath(path, -1, "virt-sensor"))
    local stateList = getStatelist(sensor)
    if stateList then
      return s.choiceType{
        _H = h,
        _P = p,
        _A = {"nil", "any", "other", unpack(stateList)},
        _D = 'any',
      }
    end

    return s.stringType{
      _H = h,
      _P = p,
      _D = 'any',
    }
  end
end

local function wc2regexp(wc)
  return '^'..wc:gsub('[%[%^%$%(%)%%%.%[%]%*%+%-%?%]]', 
  function(c) return (c=='*' and '.' or '%')..c end)..'$'
end

local function timerEnableVal(val)
  local val = s.integerVal(val, 1, math.huge) or s.booleanVal(val)
  if val~=nil then
    return val
  end
  return nil, "true, false or positive integer"
end

eventScheme = s.newScheme{
  _H = 'event',
  _Z = function(path, sch, conf)
    conf = s.addDefault(path, sch, conf, true) --clone conf with defaults
    C.updateNeteventCallback(ifStateEventTable, ifStateEvent)

    sensorState = {}
    eventHandlerTable = {}
    for _, v in ipairs(conf or {}) do
      local ev = {
        sensor = wc2regexp(v['virt-sensor']),
        prev = v['prev-state'],
        state = v.state,
        action = v.action,
      }
      eventHandlerTable[#eventHandlerTable+1] = ev
    end
    return true
  end,

  ["event-actions"] = s.newScheme{
    _H = "eventEventactions",
    tcpsender = ts.tcpsender,
  },

  ["event-generators"] = s.newScheme{
    _H = "eventEventgenerators",
    _Z = restartEventGenerators,
    scheduler = s.newScheme {
      _H = "eventEventgeneratorScheduler",
      _P = 100,
      _Z = schedulerApply,
      _LIST = s.newScheme {
        _H = "eventEventgeneratorSchedulerX",
	enable = s.booleanType{
	  _H = "eventEventgeneratorSchedulerEnable",
	  _P = 110,
	  _D = true,
	},
        ["event-name"] = s.stringType{
          _H = "eventEventgeneratorSchedulerEventname",
          _P = 100,
          _D = "alarm",
        },
        min = s.stringType{
          _H = "eventEventgeneratorSchedulerMin",
          _P = 90,
          _V = cronValidate,
          _D = '*',
        },
        hour = s.stringType{
          _H = "eventEventgeneratorSchedulerHour",
          _P = 80,
          _V = cronValidate,
          _D = '*',
        },
        day = s.stringType{
          _H = "eventEventgeneratorSchedulerDay",
          _P = 70,
          _V = cronValidate,
          _D = '*',
        },
        month = s.stringType{
          _H = "eventEventgeneratorSchedulerMonth",
          _P = 60,
          _V = cronValidate,
          _D = '*',
        },
        wday = s.stringType{
          _H = "eventEventgeneratorSchedulerWday",
          _P = 50,
          _V = cronValidate,
          _D = '*',
        },
      }
    },
    timer = s.newScheme {
      _H = "eventEventgeneratorTimer",
      _P = 100,
      _Z = timerApply,
      _NAME = s.newScheme {
        _H = "eventEventgeneratorTimerX",
        enable = s.stringType{
          _H = "eventEventgeneratorTimerEnable",
          _P = 100,
	  _D = true,
	  _V = timerEnableVal,
        },
        ["event-name"] = s.stringType{
          _H = "eventEventgeneratorTimerEventname",
          _P = 90,
          _D = "tick",
        },
        timeout = s.integerType{
          _H = "eventEventgeneratorTimerTimeout",
          _A = {1, 60 * 60 * 24 * 365},
          _D = 60,
        },
      },
      _NAME_TYPE = s.snVal,
    },

    _NAME = s.newScheme {
      _Z = restartEventGenerator,
      _H = "eventEventgeneratorX",
      enable = s.booleanType{
        _H = "eventEventgeneratorEnable",
        _D = false,
        _P = 100,
      },
      ["log"] =  s.newScheme{
        _H = "eventEventgeneratorLog",
        _P = -100,
        _X = function(path)
          path = s.clonePath(path, -1)
          io.write(s.ser(path))
          return C.getDaemonLog(path)
        end,
        _R = 'user',
      },
      ["command"] = s.stringType{
        _H = "eventEventgeneratorCommand",
      },
    }
  },

  log = s.newScheme{
    _H = 'eventLog',
    _X = function()
      return eventHandlerLogShow()
    end,
    _R = 'user',
  },
  ['simulate-event'] = RootOrAdminScheme(s.stringType{
    _H = 'eventSimulate',
    _X = function(path, sch, val)
      local sensor, event = val:match"^([^%s:]+):([^%s:]+)"
      if sensor and event then
        return eventHandler(sensor, event)
      else
        return "Bad event syntax"
      end
    end,
    _R = "user"
  }),
  _LIST = s.newScheme{
    _H = 'eventList',
    ["virt-sensor"] = s.stringType{
      _H = 'eventSensor',
      _P = 40,
      _DEP = 1,
    },
    ['prev-state'] = stateScheme('eventPrevstate', 30), -- _H = 'eventPrevstate'
    ["state"] = stateScheme('eventState', 20),          -- _H = 'eventState'
    action = s.stringType{
      _H = 'eventAction',
      _P = 10,
    },
    ["test-event"] = s.newScheme{
      _H = 'eventTestevent',
      _X = function(path)
        local conf = C.getConfByPath(s.clonePath(path, -1))
        local res, err = execScript(conf.action)
        return res or err
      end,
    }
  }
}

local function restartEventserverDaemon(path, sch, conf, prevConf)
  if not s.confCmp(conf, prevConf) then
    local conf = s.addDefault(path, sch, conf, true)
    conflib.writeToFile("/tmp/nsgsenserhttpport", tostring(conf["http-port"] or 50080))
    if conf["enable"] then
      local com = {
        "/usr/sbin/nsgsenser",
        tostring(conf["tcp-port"] or 50015),
        tostring(conf["http-port"] or 50080),
      }
      C.restartDaemon(path, com, {atExit = C.increasingSleep})
    else
    C.restartDaemon(path)
    end
  end
  return true
end

eventServerScheme = s.newScheme{
  _H = "eventEventserver",
  _HTML = "nsgsenserweb",
  _Z = restartEventserverDaemon,
  enable = s.booleanType{_H = "eventEventserverEnable"},
  ["tcp-port"] = s.tcpPortType{
    _H = "eventEventserverTcpport",
    _D = 50015,
  },
  ["http-port"] = s.tcpPortType{
    _H = "eventEventserverHttpport",
    _D = 50080,
  },
  ["run-test"] = s.newScheme{
    _H = "eventEventserverRuntest",
    _X = function(path, sch, conf, sesNum)
      local tcpport = C.getConfOrDefault(s.clonePath(path, -1, "tcp-port"))
      C.restartDaemon(path, 
      {'/usr/sbin/nsgsensertest', tostring(tcpport)},
      {sesNum = sesNum, noTimestamp = true, noRestart = true})
    end,
    _XI = true,
    _P = -99,
  },
  ["log"] =  s.newScheme{
    _H = "eventEventserverLog",
    _X = conflib.logDaemonFunc,
    _P = -100,
    _R = 'user',
  },
}

