module(..., package.seeall)
local s = require'scheme'
local socket = require'socket'
local netflow = require'nsgnetflow'
local ip = "/usr/sbin/ip"
local tc = "/usr/sbin/tc"
local grep = "/bin/grep"
local qdiscRoot = "qdisc"


local ifaceAliases = {}
local ifaceAliasesHooks = {}

ifaceAliasesGlobal = ifaceAliases

local function callHooks(alias, name)
   for _, f in pairs(ifaceAliasesHooks) do
      f(alias, name)
   end
end

function setAliasHook(k, f)
   ifaceAliasesHooks[k] = f
end

function addIfaceAlias(alias, name)
   if ifaceAliases[alias] ~= name then
      ifaceAliases[alias] = name
      if alias ~= name then
         C.log_debug('netevent: {port="'..alias..'",ifname="'..name..'"}')
         callHooks(alias, name)
      end
   end
end

function getIfaceName(alias)
   if not alias then
      return nil
   end
   local pref,suf = alias:match("^([^.]+)(.*)$")
   return (ifaceAliases[pref] or pref)..suf
end

function ifaceExist(name)
  for k, v in pairs(ifaceAliases) do
    if v == name then
      return k
    end
  end
  return false
end

function getAliasName(iface)
   local alias = ifaceExist(iface)
   local base
   while alias do
      base = alias   -- save previous
      alias = ifaceExist(alias)
   end
   return base or alias
end

local function setIfAddress(devName, ifAddress)
  if ifAddress and ifAddress.prefix then
    local peer, broadcast, anycast = '', '', ''
    if ifAddress.peer then
      peer = "peer "..ifAddress.peer
    end
    if ifAddress.broadcast then
       broadcast = "broadcast "..(ifAddress.broadcast or '+')
    end
    if ifAddress.anycast then
      anycast = "anycast "..ifAddress.anycast
    end
    local com = ("%s addr add dev %s %s %s %s %s 2>&1"):format(ip, devName,
    ifAddress.prefix, peer, broadcast, anycast)
    local res = s.readpopen(com)
    if not res:match'^%s*$' then
      C.log_warn(com..'\n'..res)
    else
      C.log_debug(com)
    end
  end
end

local function setIfAddresses(devName, ifAddress)
  if ifAddress then
    setIfAddress(devName, ifAddress)
    for _, v in ipairs(ifAddress) do
      setIfAddress(devName, v)
    end
  end
end

local devIfAddressesConfigs = {}

local applyIfAddresses =
   function(dev, flush)
      if dev and devIfAddressesConfigs[dev] then
         local cfg = devIfAddressesConfigs[dev]
         if cfg.conf and cfg.conf.configurable == "dhcp" then
            local Cmd = {"sh","-c",""}
            if cfg.conf["dhcp-options"] then
               if cfg.conf["dhcp-options"]["discard-default-gw"] == true then
                  Cmd[#Cmd] =  Cmd[#Cmd].."export discardrouter=1 ; "
               end
               if cfg.conf["dhcp-options"]["default-gw-metric"] then
                  Cmd[#Cmd] =  Cmd[#Cmd].."export metric="..tostring(cfg.conf["dhcp-options"]["default-gw-metric"].." ; ")
               end
            end

            Cmd[#Cmd] = Cmd[#Cmd].."udhcpc -f -i "..dev..";"   -- foreground
            C.restartDaemon(cfg.path, Cmd)
            C.fakeLastComTimestamp() -- user may lose connection when ip addr changed
         else
            C.restartDaemon(cfg.path)  -- kill "dhcp daemon" if not needed
            if cfg.conf and cfg.conf.configurable == "yes" then
               if flush then
                  os.execute(ip..' addr flush dev '..dev.." >/dev/null 2>&1")
               end
               setIfAddresses(dev, cfg.conf)
               C.fakeLastComTimestamp() -- user may lose connection when ip addr changed
            end
         end
      end
   end



local resetIfAddressesNeteventKey = {}
local resetIfAddressesNeteventCallBack =
   function(msg)
      if (msg.type == 'newlink') then
         applyIfAddresses(msg.ifname, false)
      end
   end

function resetIfAddresses(path, sch, conf, prevConf)
   conf = s.addDefault(path, sch, conf, true)
   local devAdm = C.getConfOrDefault(s.clonePath(path, -1, "adm-state")) or "up"
   local dev = C.getConfOrDefault(s.clonePath(path, -1, "ifName"))  or getIfaceName(path[#path - 1])
   if devAdm=="up" then
      --save current config to global table
      devIfAddressesConfigs[dev] = {
         conf = conf,
         path = s.deepcopy(path),
      }
      C.updateNeteventCallback( -- try set on dev creation
         resetIfAddressesNeteventKey,
         resetIfAddressesNeteventCallBack
      )
      applyIfAddresses(dev, true) -- try set now
   else
      devIfAddressesConfigs[dev] = nil
   end
   return true
end

local qdiscPrioDefMap = {1, 2, 2, 2, 1, 2, 0, 0 , 1, 1, 1, 1, 1, 1, 1, 1}

local qdiscCriterionPriority = {
   pfifo_fast = {},
   pfifo = {
      "limit"
   },
   bfifo = {
      "limit"
   },
   red = {
      "limit",
      "min",
      "max",
      "avpkt",
      "burst",
      "probability",
      "bandwidth",
      "ecn",
   },
   sfq = {
  --    "limit",
      "perturb",
      "quantum",
   },
   tbf = {
      "limit",
      "burst",
      "rate",
      "mtu",
      "peakrate",
      "latency",
   },
   htb = {
      "default"
   },
   prio = {
      "bands",
      "priomap",
   },
}

local  qdiscCriterionPatterns = {
   pfifo = {
      limit = "limit %d"
   },
   bfifo = {
      limit = "limit %d"
   },
   red = {
      limit = "limit %d",
      min = "min %d",
      max = "max %d",
      avpkt = "avpkt %d",
      burst = "burst %d",
      probability = "probability %d",
      bandwidth = "bandwidth %d",
      ecn = "ecn",
   },
   sfq ={
   --   limit = "limit %d",
      perturb = "perturb %d",
      quantum = "quantum %d",
   },
   tbf = {
      limit = "limit %d",
      burst = "burst %s",
      rate = "rate %d",
      mtu = "mtu %s",
      peakrate = "peakrate %d",
      latency  = "latency %d",
   },
   htb = {
      default = "default %s"
   },
   prio = {
      bands = "bands %d",
      priomap = function(cfg)
                   local map = {"priomap"}
                   local i = 0
                   while i < 16 do
                      map[#map + 1] = cfg[i] or qdiscPrioDefMap[i+1]
                      i = i + 1
                   end
                   return table.concat(map, ' ')
                end,
   },
}

local qdiscClassesCriterionPriority = {
   htb = {
      "burst",
      "cburst",
      "ceil",
      "rate"
   },
}

local qdiscClassesCriterionPatterns = {
   htb = {
     burst = "burst %d",
     cburst = "cburst %d",
     ceil = "ceil %d",
     rate = "rate %d"
   },
}
local applyQdisc
local applyClass
local getHandle

function getHandle(handles, default)
   if default and not handles[default] then
      handles[default] = true
      return default
   else
      local i = 1
      while handles[i] do
         i = i + 1
      end
      handles[i] = true
      return i
   end
end

function applyClass(qdisc, dev, conf, hiParent, loParent, handles, err, log)
   for k,v in pairs(conf.classes or {}) do
      if (qdisc ~= "prio") then --prio has autocreated classes
         cmd = {
            tc, "class", "add", "dev", dev,
            ("parent %s:%s"):format(hiParent, loParent),
            ("classid %s:%s"):format(hiParent,k ),
            qdisc
         }
         for _, crit in ipairs(qdiscClassesCriterionPriority[qdisc] or {}) do
            if v and v[crit] then
               cmd[#cmd + 1] = (qdiscClassesCriterionPatterns[qdisc][crit]
                             ):format(v[crit])
            end
         end
         local res = s.readpopen(table.concat(cmd," ").." 2>&1")
         if not res:match'^%s*$' then
            err[#err + 1] = "Error in cmd:"..table.concat(cmd," ").."\n"..res
            C.log_error(err[#err])
         else
            C.log_debug(table.concat(cmd," "))
            log[#log + 1] = table.concat(cmd," ")
         end
      end
      if v["queue-type"] == "classes" then
         applyClass(qdisc, dev, v, hiParent, k, handles, err, log)
      else
         applyQdisc(dev, v, hiParent..":"..k, handles, err, log)
      end
   end
end

function applyQdisc(dev, conf, parent, handles, err, log)
   local qdisc = conf["queue-type"]
   local handle = conf.handle or getHandle(handles)

   if parent == 'root' then
      cmd = {
         tc, "qdisc", "del", "dev", dev, parent
      }
      os.execute(table.concat(cmd," ").." 1>/dev/null 2>/dev/null")
      C.log_debug(table.concat(cmd," "))
      log[#log + 1] = table.concat(cmd," ")
   end

   if qdisc == "pfifo_fast" then
      return
   end

   conf = conf[qdisc]
   cmd = {
      tc, "qdisc", "add", "dev", dev, (parent=="root" and "root" or "parent "..parent),
      ("handle %s:"):format(handle),
      qdisc,
   }
   for _, crit in ipairs(qdiscCriterionPriority[qdisc] or {}) do
      if conf and conf[crit]
      then
         do
            if (type(qdiscCriterionPatterns[qdisc][crit])== "function")
            then
               cmd[#cmd + 1] = qdiscCriterionPatterns[qdisc][crit](conf[crit])
            else
               cmd[#cmd + 1] = (qdiscCriterionPatterns[qdisc][crit]
                             ):format(conf[crit])
            end
         end
      end
   end
   local res = s.readpopen(table.concat(cmd," ").." 2>&1")
   if not res:match'^%s*$' then
      err[#err + 1] = "Error in cmd:"..table.concat(cmd," ").."\n"..res
      C.log_error(err[#err])
   else
      C.log_debug(table.concat(cmd," "))
      log[#log + 1] = table.concat(cmd," ")
   end

   if (qdisc == "htb") or (qdisc == "prio") then
      applyClass(qdisc, dev, conf, handle, 0, handles, err, log)
   end
end

local qdiscFindHandles
qdiscFindHandles =
   function(handles, cfg, path, selfPath)
      if cfg and cfg.handle and (selfPath ~= path..".handle") then
         handles[tonumber(cfg.handle)] = true
      end
      for k,v in pairs(cfg and cfg[cfg["queue-type"]] and cfg[cfg["queue-type"]].classes or cfg.classes or {}) do
         if type(tonumber(k)) == "number" then
            qdiscFindHandles(handles, v, path..(cfg[cfg["queue-type"]] and cfg[cfg["queue-type"]].classes and "."..cfg["queue-type"] or '')..".classes."..k, selfPath)
         end
      end
   end

local devIpLinkConfigs = {}

local applyDescription =
   function (dev)
      if dev and devIpLinkConfigs[dev] and devIpLinkConfigs[dev].descr then
         local descr = devIpLinkConfigs[dev].descr
         descr = descr and descr ~= "" and descr or "'Configured by NSG.'"
         descr = tostring(descr):sub(1,250):gsub("\\","\\\\"):gsub("'","\\'"):gsub("`","\\`"):gsub('"','\\"'):gsub('[^%w%p%s]','')
         os.execute("ip link set "..dev.." alias "..descr)
      end
   end

function applyIpLink(dev)
   local conf = devIpLinkConfigs[dev] and devIpLinkConfigs[dev].conf
   if conf then
      local com = {
         ip, 'link', 'set', dev,
      }
      if conf.arp then
         com[#com+1] = 'arp'
         com[#com+1] = conf.arp
      end
      if conf.multicast then
         com[#com+1] = 'multicast'
         com[#com+1] = conf.multicast
      end
      if conf.dynamic then
         com[#com+1] = 'dynamic'
         com[#com+1] = conf.dynamic
      end
      if conf.qlen then
         com[#com+1] = 'txqueuelen'
         com[#com+1] = conf.qlen
      end
      if conf.mtu then
         com[#com+1] = 'mtu'
         com[#com+1] = conf.mtu
      end
      if conf['mac-address']
      then -- don`t change mac if it exists
         if os.execute(ip..' link ls dev '..dev.." | "..grep.." -oie '"..conf['mac-address'].."'") ~= 0 then
            com[#com+1] = 'address'
            com[#com+1] = conf['mac-address']
            if os.execute(ip.." link ls dev "..dev.." | "..grep.." -oie '[^[:alnum:]]UP[^[:alnum:]]'") == 0 then
               C.log_debug(ip.." link set dev "..dev.." down")
               os.execute(ip.." link set dev "..dev.." down")
            end
            com[#com+1] = devIpLinkConfigs[dev]["adm-state"]
         end
      end
      com[#com+1] = "2>&1"

      if #com>5 then --may be nothing to do ?
         local res = s.readpopen(table.concat(com,' '))
         if not res:match'^%s*$' then
            C.log_warn(table.concat(com,' ')..'\n'..res)
         else
            C.log_debug(table.concat(com,' '))
         end
      end
      if conf.qdisc then
         local err = {}
         local log = {}
         local handles = {}
         qdiscFindHandles(handles, conf.qdisc, "", nil)
         applyQdisc(dev, conf.qdisc, "root", handles, err, log)
         devIpLinkConfigs[dev].tcLog = table.concat(log,"\n")
         if #err then
            return nil, table.concat(err,"\n")
         end
      end
   end
   return true
end

local resetIpLinkNeteventKey = {}
local resetIpLinkNeteventCallBack =
   function(msg)
      if (msg.type == 'newlink') then
         applyIpLink(msg.ifname)
         applyDescription(msg.ifname)
      end
   end

function ipLinkSet(path, sch, conf, prevConf)
   conf = s.addDefault(path, sch, conf, true) --clone conf with defaults
   local devAdm = C.getConfOrDefault(s.clonePath(path, -1, "adm-state")) or "up"
   local dev = C.getConfOrDefault(s.clonePath(path, -1, "ifName"))  or getIfaceName(path[#path - 1]) 
   
   if devAdm=="up" then
      --save current config to global table
      devIpLinkConfigs[dev] = {
         conf = conf,
         descr =  C.getConfOrDefault(s.clonePath(path, -1, "description")),
         ["adm-state"] = C.getConfOrDefault(s.clonePath(path, -1, "adm-state")),
      }
      C.updateNeteventCallback( -- try set on dev creation
         resetIpLinkNeteventKey,
         resetIpLinkNeteventCallBack
      )
      applyIpLink(dev) -- try set now
      applyDescription(dev)
   else
      devIpLinkConfigs[dev] = nil
   end
  return true
end

local qdiscScheme
local qdiscClassesScheme

qdiscClassesScheme =
   function(_, path)
      local instances= {
         htb = {"pfifo_fast", "pfifo", "bfifo", "red", "sfq" ,"tbf", "prio", "classes"},
         prio = {"pfifo_fast", "pfifo", "bfifo", "red", "sfq" ,"tbf", "prio", "htb"},
      }

      local instanceScheme = s.choiceType{
         _H = "qdiscType",
         _D = "pfifo",
         _DEP = 1,
         _P = -100,
      }

      local qdisc = 'UNKNOWN'
      local qdiscTop = 2
      local instance
      while qdiscTop < #path do
         instance = C.getConfByPath(s.clonePath(path, -qdiscTop, "queue-type" ))
         if (type(instance) == "string") and instance ~= "classes" then
            qdisc = instance
               break
         end
         qdiscTop = qdiscTop + 1
      end
      local classInstance = C.getConfByPath(s.clonePath(path, nil, "queue-type" )) or "pfifo"

      local qdiscClasses = {
         htb = s.newScheme{
            _P = -110,
            _H = "qdiscHtbClasses",
            _NAME_TYPE =
               function (val)
                  if not s.integerVal(val, 1, 65535) then
                     return s.integerVal(val, 1, 65535)
                  end
                  local selfPath = table.concat(path,'.').."."..val
                  local topPath = s.clonePath(path, -qdiscTop, "htb") or {}
                  local cfg =  C.getConfByPath(topPath) or {}
                  local findClasses
                  findClasses =
                     function(classes,cfg, path)
                        for k,v in pairs(cfg and cfg.classes or {}) do
                           if (type(tonumber(k)) == "number")
                           and (selfPath ~= path..".classes."..k) then
                              classes[k] = true
                              findClasses(classes, v, path..".classes."..k)
                           end
                        end
                     end
                  local classes = {}
                  findClasses(classes, cfg, table.concat(topPath,"."))
                  if classes[val] then
                     return nil, ("Class %s already exist."):format(tostring(val))
                  else
                     return s.integerVal(val, 1, 65535)
                  end
               end,
            _NAME =
               function (_, path)
                  local  instance = C.getConfByPath(s.clonePath(path, nil, "instance" )) or "pfifo"
                  local scheme = s.newScheme{
                     _H = "qdiscHtbClass",
                     rate = s.integerType{_H = "qdiscHtbRate"},
                     ceil = s.integerType{_H = "qdiscHtbCeil"},
                     burst = s.integerType{_H = "qdiscHtbBurst"},
                     cburst = s.integerType{_H = "qdiscHtbCburst"},
                     ["queue-type"] = instances[qdisc] and instanceScheme{_A = instances[qdisc]},
                  }

                  if instance=='classes' then
                     return scheme{classes = qdiscClassesScheme,}
                  else
                     return qdiscScheme(_, path)(scheme)
                  end
               end,
         },
         prio = function(_, path)
                   local scheme = s.newScheme{_H = "qdiscPrioClasses", _P = -110,}
                   local classScheme =
                      function (_, path)
                         local  instance = C.getConfByPath(s.clonePath(path, nil, "queue-type" )) or "pfifo"
                         local sch = s.newScheme{
                            _H = "qdiscPrioClass",
                            ["priomap-index"] = s.constantType{
                               _H = "qdiscPrioClassIndex",
                               _D = path[#path] - 1,
                            },
                            ["queue-type"] = instances[qdisc] and instanceScheme{_A = instances[qdisc]},
                         }
                         return qdiscScheme(_, path)(sch)
                      end
                   local bands = C.getConfByPath(s.clonePath(path, -1, "bands") or {}) or 3
                   while bands > 0 do
                      scheme = scheme{ [bands] = classScheme }
                      bands = bands - 1
                   end
                   return scheme
                end,
         UNKNOWN = {},
      }

      if type(qdiscClasses[qdisc]) == 'function' then
         return qdiscClasses[qdisc](_,path)
      else
         return qdiscClasses[qdisc]
      end
   end

local qdiscPriomapScheme
qdiscPriomapScheme =
   function(_, path)
      local bands = C.getConfOrDefault(s.clonePath(path, -2, "bands") or {}) or 3
      local _A = {}
      local i = 0
      while i < bands do
         _A[#_A + 1] = i
         i = i + 1
      end
      scheme = s.choiceType{
         _H = "qdiscPrioPriomapItem",
         _A = _A,
         _D = path and #path>0 and tonumber(path[#path]) and  type(qdiscPrioDefMap[path[#path] + 1])=="number" and  qdiscPrioDefMap[path[#path] + 1] < bands and qdiscPrioDefMap[path[#path] + 1] or 0,
      }
      return scheme
   end

qdiscScheme =
   function(_, path)
      local conf = C.getConfByPath(s.clonePath(path,-2))
      if conf and ((conf["bridge-group"]) or (conf["bond-group"])) then
         return nil
      end
      local qdiscTypeScheme = {
         pfifo_fast = {_H = "qdiscPfifofast", },
         pfifo = {_H = "qdiscPfifo", limit = s.integerType{_H = "qdiscPfifoLimit"}},
         bfifo = {_H = "qdiscBfifo", limit = s.integerType{_H = "qdiscBfifoLimit"}},
         red = {_H = "qdiscRed",
                limit = s.integerType{_H = "qdiscRedLimit"},
                min = s.integerType{_H = "qdiscRedMin"},
                max = s.integerType{_H = "qdiscRedMax"},
                avpkt = s.integerType{_H = "qdiscRedAvpkt"},
                burst = s.integerType{_H = "qdiscRedBurst"},
                ecn = s.booleanType{_H = "qdiscRedEcn"},
                bandwidth = s.integerType{_H = "qdiscRedBandwidth"},
                probability = s.integerType{_H = "qdiscRedProbability"},},
         sfq = {_H = "qdiscSfq",
            --    limit = s.integerType{_H = "qdiscSfqPerturb"},
                perturb = s.integerType{_H = "qdiscSfqPerturb"},
                quantum = s.integerType{_H = "qdiscSfqQuantum"},
             },
         tbf = {_H = "qdiscTbf",
                _N = function(sch, conf)
                        if conf and conf.latency then
                           return {nil, limit = true}
                        elseif  conf and conf.limit then
                           return  {nil, latency = true}
                        end
                        return {}
                     end,
                limit = s.integerType{_H = "qdiscTbfLimit", _DEP = 1},
                latency = s.integerType{_H = "qdiscTbfLatency", _DEP = 1},
                rate = s.integerType{_H = "qdiscTbfRate"},
                burst = s.variantType{_H = "qdiscTbfBurst",
                                      _A = {s.integerType{_A = {1,2^32}},s.stringType{}}},
                --mpu = s.integerType{_H = "qdiscTbfMpu"},
                peakrate = s.integerType{_H = "qdiscTbfPeakrate"},
                mtu = s.anyType{_H = "qdiscTbfMtu"},
             },
         htb = {_H = "qdiscHtb",
                default =  s.integerType{_H = "qdiscHtbDefault", _P = 50,},
                ["queue-type"] = s.constantType{_H = "qdiscInstance", _D = "classes", _A = {"classes"}, _P = 30},
                classes = qdiscClassesScheme,
             },
         prio = {_H = "qdiscPrio",
                 bands =  s.integerType{_H = "qdiscPrioBands",
                                        _P = 50,
                                        _DEP = 1,
                                        _D = 3,
                                        _A = {1,64},
                                     },
                 ["queue-type"] = s.constantType{_H = "qdiscInstance", _D = "classes", _P = 30},
                 classes = qdiscClassesScheme,
                 priomap = s.newScheme{
                    _H = "qdiscPrioPriomap",
                    _NAME = qdiscPriomapScheme,
                    _NAME_TYPE = s.integerVal,
                    _A  = {0,15},
                 },
              },
      }

      instances = {
         [true] = {"pfifo_fast", "pfifo", "bfifo", "red", "sfq" ,"tbf", "htb", "prio"},
         [false] = {"pfifo_fast", "pfifo", "bfifo", "red", "sfq" ,"tbf", "prio"}
      }

      local isRoot =  (path[#path] == qdiscRoot)
      local defInstance = (isRoot and "pfifo_fast") or "pfifo"
      local instance = C.getConfByPath(s.clonePath(path, nil, "queue-type" )) or  defInstance

      local instanceScheme = s.choiceType{
         _H = "qdiscType",
         _A = instances[isRoot],
         _D = defInstance,
         _DEP = 1,
         _P = -100,
      }

      local sch1 = s.newScheme{
         _H = "qdisc",
         handle = s.integerType{
            _P = -105,
            _H = "qdiscHandle",
            _V = function(val)
                    local topPath =  s.clonePath(path)
                    local selfPath =  table.concat(s.clonePath(path, nil, "handle"),".")

                    while topPath and (topPath[#topPath] ~= qdiscRoot) do
                       topPath[#topPath] = nil
                    end

                    if #topPath == 0 then
                       return nil, "Can`t find root point of qdisc."
                    end
                    local cfg = C.getConfByPath(topPath) or {}

                    topPath = table.concat(topPath, ".")

                    local handles = {}
                    qdiscFindHandles(handles, cfg, topPath, selfPath)
                    if handles[tonumber(val)] then
                       return nil, "Handle already exist! SelfPath:"..selfPath
                    else
                       return s.integerVal(val, 1, 65535)
                    end
                 end,
         },
         ["queue-type"] = instanceScheme,
         [instance] = (instance == 'classes' and qdiscClassesScheme) or s.newScheme(qdiscTypeScheme[instance])({_P = -110}),
      }
      if isRoot then
         sch1 = sch1{
            show = s.newScheme{
               _H = "ifShow",
               _X = function(path)
                       local dev = getIfaceName(path[#path-3])
                       return (
                          devIpLinkConfigs[dev].tcLog or "No tc log\n"
                       ).."\n\n"..s.readpopen(
                          table.concat(
                             { tc,
                               "-s qdisc ls dev",
                               dev,
                               "&&",
                               tc,
                               "-s class ls dev",
                               dev
                            },
                             ' ')
                       )
                 end,
               _R = 'user',
            },
         }
      end

      return sch1
   end

local ifAddressScheme = s.newScheme{
   _H = "interface address",
   prefix =  function(_, path)
                local dhcp = C.getConfByPath(s.clonePath(path, -1, "configurable" ))
                return (dhcp ~= "dhcp")
                and s.ip46PrefixType{_H = "ifAddressPrefix", _P = 100, _DEP = 1}
             end,
   peer =
      function(_, path)
         local dhcp = C.getConfByPath(s.clonePath(path, -1, "configurable" )) ~= "dhcp"
         local prefix = dhcp and C.getConfByPath(s.clonePath(path, -1, "prefix" )) or nil
         return prefix and s.ipprefixVal(prefix) and s.ip4AddressType{_H = "ifAddressPeer", _P = 90} or nil
           end,
   broadcast =
      function(_, path)
         local dhcp = C.getConfByPath(s.clonePath(path, -1, "configurable" )) ~= "dhcp"
         local prefix = dhcp and C.getConfByPath(s.clonePath(path, -1, "prefix" ))
         return prefix and s.ipprefixVal(prefix) and s.ip4AddressType{_H = "ifAddressBroadcast", _P = 80} or nil
            end,
   anycast =
      function(_, path)
         local dhcp = C.getConfByPath(s.clonePath(path, -1, "configurable" )) ~= "dhcp"
         local prefix = dhcp and C.getConfByPath(s.clonePath(path, -1, "prefix" ))
         return prefix and s.ipprefixVal(prefix) and s.ip4AddressType{_H = "ifAddressAnycast", _P = 70} or nil
      end,
}

ifAddressListScheme = ifAddressScheme{
  _H = "ifAddress",
  _Z = resetIfAddresses,
  configurable =  s.newScheme{
    _T = "choice",
    _A = {
      {"yes"},
      {"no"},
      {"dhcp"},
    },
    _V = s.choiceVal,
    _H = "ifAddressEnable",
    _D = "yes",
    _P = 15,
    _DEP = 1,
  },
  ["dhcp-options"] =
     function(_, path)
        local dhcp = C.getConfByPath(s.clonePath(path, -1, "configurable" ))
        if dhcp == 'dhcp' then
           return s.newScheme{
              _P = 10,
              _H = "ifaceDhcpOptions",
              ["discard-default-gw"] = s.booleanType{
                 _H = "ifaceDiscardDefaultgw",
                 _D = false,
                 },
              ["default-gw-metric"] = s.integerType{_H = "ifaceDefaultgwMetric", _A={0,32767}},
           }
        else
           return nil
        end
     end,
  _LIST = function(_, path)
             local dhcp = C.getConfByPath(s.clonePath(path, -1, "configurable" ))
             return (dhcp ~= "dhcp") and ifAddressScheme{
                _H = "ifAddressSecondary",
             }
          end,
}

linkScheme = s.newScheme{
  _H = "ifLink",
  _Z = ipLinkSet,
  arp =  s.newScheme{
    _T = "choice",
    _A = {"on", "off"},
    _V = s.choiceVal,
    _H = "ifLinkArp",
  },
  multicast =  s.newScheme{
    _T = "choice",
    _A = {"on", "off"},
    _V = s.choiceVal,
    _H = "ifLinkMulticast",
  },
  dynamic =  s.newScheme{
    _T = "choice",
    _A = {"on", "off"},
    _V = s.choiceVal,
    _H = "ifLinkDynamic",
  },
  qlen = function(_, path)
              local conf = C.getConfByPath(s.clonePath(path,-2))
              if conf and ((conf["bridge-group"]) or (conf["bond-group"])) then
                 return nil
              end 
              return s.newScheme{
                 _T = "number",
                 _A = {0, 2147483647},
                 _V = s.integerVal,
                 _H = "ifLinkTxlen",
              } 
           end,
  qdisc = qdiscScheme,
  mtu =  s.newScheme{
    _T = "number",
    _A = {0, 9000},
    _V = s.integerVal,
    _H = "ifLinkMtu",
  },
  ['mac-address'] = s.newScheme{
    _T = 'string',
    _V = s.macaddressVal,
    _H = "ifLinkMacaddress",
  },
}

function showInterfaceStat(iface)
   local addr = s.readpopen(ip..' addr show dev '..iface.." 2>&1")
   local link = s.readpopen(ip..' -s -s link show dev '..iface.." 2>&1")
   alias = ((link:match("\n[%s]+alias[%s]+.-\n") or ''):match("alias[%s]+.-\n") or ""):gsub("alias", "Description:")
   link = link:gsub("^.-\n(%s+RX%:.*)$", "%1")
   return "Interface name:"..iface.."\n"..addr..(alias or '')..link
end


function interfaceMainApply(path, sch, conf)
   local devName = C.getConfOrDefault(s.clonePath(path, nil, "ifName")) or getIfaceName(path[#path])
   local adm = C.getConfOrDefault(s.clonePath(path, nil, "adm-state"))
   -- (conf or {})["adm-state"] or sch["adm-state"]._D or "down"
   local com = {ip, 'link', 'set', devName, adm, "2>&1"}
   local res = s.readpopen(table.concat(com,' '))
   if not res:match'^%s*$' then
      C.log_warn(table.concat(com,' ')..'\n'..res)
   else
      C.log_debug(table.concat(com,' '))
   end

   netflow.netflows[s.ser(path)] = (conf or {}).netflow
   netflow.restartNetflow(s.ser(path))

   return true
end

interfaceScheme = s.newScheme{
   _H = "if1",
   _Z = function(path, sch, conf)
           return interfaceMainApply(path, sch, conf)
        end,
   ["adm-state"] =  s.newScheme{
      _T = "choice",
      _A = {"up", "down"},
      _V = s.choiceVal,
      _H = "ifLinkAdm",
      _D = "up"
   },
   ifAddress = function(_, path)
                  local conf = C.getConfByPath(s.clonePath(path,-1))
                  if conf and ((conf["bridge-group"]) or (conf["bond-group"])) then
                     return nil
                  end
                  return ifAddressListScheme
               end,
   description = s.stringType{_H = "ifDescription", _P = 0},
   link = linkScheme,
   show = s.newScheme{
      _H = "ifShow",
      _X = function(path)
              return "Port name:"..tostring(path[#path-1]).."\n"..showInterfaceStat(getIfaceName(path[#path-1]))
           end,
      _R = 'user',
   },
   netflow = s.integerType{
      _H = "ifNetflow",
      _A = {1,255}
   }
}

local LO_PREFIX = "127.0.0.1/8"
local loInterfaceScheme = s.newScheme{
  _H = "ifLo",
  _Z = function() return true end,
  ifAddress = s.newScheme{
    _H = "ifAddress",
    _Z = resetIfAddresses,
    prefix = s.newScheme{
      _T = "string",
      _V = function(val)
        return val == LO_PREFIX and LO_PREFIX or
        nil, LO_PREFIX
      end,
      _H = "ifAddressPrefix",
      _D = LO_PREFIX,
    },
    configurable =  s.newScheme{
      _T = "choice",
      _H = "ifAddressEnable",
      _A = {"yes", "no", "dhcp"},
      _V = s.choiceVal,
      _D = "no",
    },
    _LIST = ifAddressScheme{
      _H = "ifAddressSecondary",

    },
  },
  show = s.newScheme{
    _H = "ifShow",
    _X = function(path)
      return "Port name:"..tostring(path[#path-1]).."\n"..showInterfaceStat(path[#path-1])
    end,
    _R = 'user',
  },
}

interfaceListScheme = s.newScheme{
  _H = "if0",
  _Z = function() return true end,
  lo = loInterfaceScheme,

  dummy0 = interfaceScheme{
    _H = "ifDummy",
    _Z = function(path, sch, conf) 
      os.execute"modprobe dummy"
      return interfaceMainApply(path, sch, conf)
    end,
  },

  _NAME = interfaceScheme{
    _Z = function(path, sch, conf) 
      local alias = ifaceExist(path[#path])
      if alias then
        C.setConfigValue(path, nil)
        return false, "There is node "..alias.." used this ip interface" 
      end
      return interfaceMainApply(path, sch, conf)
    end,
  },

  _NAME_TYPE = function(val)
    local res, err =  s.stringVal(val, 16)
    if not res then
      return res, err
    end
    local alias = ifaceExist(val)
    if alias then
      return false, "interface name except lo, "..alias..'...'
    end
    return val
  end,
}

