module(..., package.seeall)
local s = require'scheme'
local ip = require'nsgiplib'
local iface_mod = require'nsgifaceconf'
local conflib = require'nsgconflib'

local applyNFNetflow

local getTable =
   function(path)
      if path[1] == "virtual-router" then
         return path[4]
      else
         return path[2]
      end
   end

local getChain =
   function(path)
      if path[1] == "virtual-router" then
         return path[5]
      else
         return path[3]
      end
   end

local xtables = {
   iptables = ip.iptables,
   icmp = {
      ip = 'icmp',
      ipv6 = 'icmpv6'
   },
   ICMP = {
      ip = 'ICMP',
      ipv6 = 'ICMPv6'
   },
   icmpproto = {
      icmp = 'ip',
      ['ipv6-icmp'] = 'ipv6'
   }
}

local chainNameVal = function (val)
  if type(val) == 'string' and val:match("^[%l][%w%p]*$") then
    return val
  else
    return nil, "word with alpha lower-case first item"
  end
end

local withInverted = function (t)
  local out = s.deepcopy(t)
  for _,v in pairs(t) do
    out[#out +1] = "! "..tostring(v)
  end
  return out
end

function genCheckValOrInverse(checkFunc)
  return function(val)
    if type(checkFunc(val)) == 'nil' and type(val)=='string' then
      local nval = val:gsub('^[%s]*![%s]*','')
      if type(checkFunc(nval)) == 'nil' then
        local _ , res = checkFunc(nval)
        res = res.." Sign '!' at start may be used for inversion."
      else
        return "! "..checkFunc(nval)
      end
    else
      return checkFunc(val)
    end
  end
end

local function getProtocols()
  local protocols = conflib.getFile"/etc/protocols"
  protocols = '\n'..(protocols or '')
  local res = {'all'}
  for proto in protocols:gmatch"\n%s*([^%s#]+)%s+%d+%s+" do
    res[#res+1] = proto
    res[#res+1] = '!'..proto
  end
  return res
end


local function getIcmpType(path)
   local help = s.readpopen(xtables.iptables[path[1]].." -p "..xtables.icmp[path[1]].." -h")
  local res = {}
  for it in help:match("Valid "..xtables.ICMP[path[1]].." Types:(.*)$"):gmatch"[%w%-]+" do
    res[#res+1] = it
    res[#res+1] = '!'..it
  end
  return res
end

local criterionPriority = {
   "protocol",
   "protocol-num",
   "mac-source",
   "tcp-flags",
   "source",
   "destination",
   "in-interface",
   "out-interface",
   "source-port",
   "destination-port",
   "icmp-type",
   "icmpv6-type",
   "explicit-matches",
   "extra-options",
   "target",
   "chain",
   "to-source",
   "to-destination",

   "set-class",
   "new",
   "hashmode",
   "clustermac",
   "total-nodes",
   "local-node",
   "hash-init",
   "connmark-params",
   "random",
   "set-dscp",
   "set-dscp-class",
   "ecn-tcp-remove",
   "queue-num",
   "log-level",
   "log-prefix",
   "log-tcp-sequence",
   "log-tcp-options",
   "log-ip-options",
   "log-uid",
   "set-mark",
   "set-xmark",
   "to-ports",
   "to",
   "rateest-name",
   "rateest-interval",
   "rateest-ewmalog",
   "reject-with",
   "set-mss",
   "clamp-mss-to-pmtu",
   "strip-options",
   "set-tos",
   "ttl-set",
   "ttl-dec",
   "ttl-inc",
   "ulog-nlgroup",
   "ulog-prefix",
   "ulog-cprange",
   "ulog-qthreshold",

   "extra-target",
}
local criterionPatterns = {
  protocol = "-p %s",
  ["protocol-num"] = "-p %d",
  ["mac-source"] = "--mac-source %s",
  source = "-s %s",
  destination = "-d %s",
  ["in-interface"] = "-i %s",
  ["out-interface"] = "-o %s",
  ["source-port"] = "--sport %s",
  ["destination-port"] = "--dport %s",
  ["tcp-flags"] = "--tcp-flags %s",
  ["icmp-type"] = "--icmp-type %s",
  ["icmpv6-type"] = "--icmpv6-type %s",
  ["extra-options"] = "%s",
  target = "-j %s",
  chain = "-g %s",
  ["to-source"] = "--to-source %s",
  ["to-destination"] = "--to-destination %s",

  ["set-class"] = "--set-class %s",
  ["new"] = "--new",
  ["hashmode"] = "--hashmode %s",
  ["clustermac"] = "--clustermac %s",
  ["total-nodes"] = "--total-nodes %d",
  ["local-node"] = "--local-node %d",
  ["hash-init"] = "--hash-init %d",
  ["connmark-params"] = "%s",
  ["random"] = "--random",
  ["set-dscp"] = "--set-dscp %d",
  ["set-dscp-class"] = "--set-dscp-class %s" ,
  ["ecn-tcp-remove"] = "--ecn-tcp-remove",
  ["queue-num"] = "--queue-num %d",
  ["log-level"] = "--log-level %d",
  ["log-prefix"] = "--log-prefix %s",
  ["log-tcp-sequence"] = "--log-tcp-sequence",
  ["log-tcp-options"] = "--log-tcp-options",
  ["log-ip-options"] = "--log-ip-options",
  ["log-uid"] = "--log-uid",
  ["set-mark"] = "--set-mark %s",
  ["set-xmark"] = "--set-xmark %s",
  ["to-ports"] = "--to-ports %s",
  ["to"] = "--to %s",
  ["rateest-name"] = "--rateest-name %s",
  ["rateest-interval"] = "--rateest-interval %s",
  ["rateest-ewmalog"] = "--rateest-ewmalog %d",
  ["reject-with"] = "--reject-with %s",
  ["set-mss"] = "--set-mss %d",
  ["clamp-mss-to-pmtu"] = "--clamp-mss-to-pmtu",
  ["strip-options"] = "--strip-options %s",
  ["set-tos"] = "--set-tos %s",
  ["ttl-set"] = "--ttl-set %d",
  ["ttl-dec"] = "--ttl-dec %d",
  ["ttl-inc"] = "--ttl-inc %d",
  ["ulog-nlgroup"] = "--ulog-nlgroup %d",
  ["ulog-prefix"] = "--ulog-prefix %s",
  ["ulog-cprange"] = "--ulog-cprange %d",
  ["ulog-qthreshold"] = "--ulog-qthreshold %d",

  ["extra-target"] = "%s",
}

local explicitModulesParams = {
  addrtype = { params = "%s"},
  ah = { params = "%s"},
  comment = { comment = "--comment '%s'"},
  connbytes = { params = "%s"},
  connlimit = { params = "%s"},
  connmark = { params = "%s"},
  conntrack = { params = "%s"},
  dscp = { params = "%s"},
  ecn = { params = "%s"},
  esp = { params = "%s"},
  hashlimit = { params = "%s"},
  helper = { params = "%s"},
  iprange = { params = "%s"},
  length = { params = "%s"},
  limit = { params = "%s"},
  mac = { ["mac-source"] = "--mac-source %s"},
  mark = { mark = "--mark %s"},
  multiport = {
    ["source-ports"] = "--source-ports %s" ,
    ["destination-ports"] = "--destination-ports %s" ,
  },
  --owner
  physdev = { ["physdev-in"] = "--physdev-in %s",
    ["physdev-out"] =  "--physdev-out %s",
    ["physdev-is-in"] = "--physdev-is-in",
    ["physdev-is-out"] =  "--physdev-is-out",
    ["physdev-is-bridged"] = "--physdev-is-bridged"
  },
  pkttype = {["pkt-type"] = "--pkt-type %s"},
  policy ={
    dir = "--dir %s",
    pol = "--pol %s",
    strict = "--strict",
    reqid = "--reqid %s",
    spi = "--spi %s",
    proto = "--proto %s",
    mode = "--mode %s",
    ["tunnel-src"] = "--tunnel-src %s",
    ["tunnel-dst"] = "--tunnel-dst %s",
    next = "--next",
  },
  quota = {quota = "--quota %d"},
  rateest = { params = "%s"},
  realm = { params = "%s"},
  recent = { params = "%s"},
  sctp = { params = "%s"},
  set = { params = "%s"},
  state = { params = "%s"},
  statistic = { params = "%s"},
  string = { params = "%s"},
  tcpmss = { params = "%s"},
  time = { params = "%s"},
  tos = { params = "%s"},
  ttl = { params = "%s"},
  u32 = { value = "--u32 '%s'"},
  unclean = { value = ""},
}

local function splitVal(val)
  if (tostring(val)):sub(1,1) == '!' then
    return val:sub(2), '! '
  else
    return val, ""
  end
end
local helpersModules={
   DCCP = {"nf_contrack_proto_dccp","nf_nat_proto_dccp"},
   SCTP = {"nf_contrack_proto_sctp","nf_nat_proto_sctp"},
   ["UDP-Lite"] = {"nf_contrack_proto_udplite","nf_nat_proto_udplite"},
   Amanda ={ "nf_contrack_amanda","nf_nat_amanda"},
   FTP = {"nf_contrack_ftp","nf_nat_ftp"},
   ["H.323"] = {"nf_contrack_h323","nf_nat_h323"},
   IRC = {"nf_contrack_irc","nf_nat_irc"},
   NetBIOS ={ "nf_conntrack_netbios_ns"},
   PPTP = {"nf_contrack_pptp","nf_nat_pptp"},
   SANE = {"nf_contrack_sane","nf_nat_sane"},
   SIP = {"nf_contrack_sip","nf_nat_sip"},
   TFTP = {"nf_contrack_tftp","nf_nat_tftp"},
   SNMP = {"nf_contrack_snmp_basic","nf_nat_snmp_basic"},
}

local function insKernelModules()
   local helpers = C.getConfByPath({"ip","nat","HELPERS"}) or {}
   local modprobe = {}
   for k,v in pairs(helpers) do
      if v and type(v)=='boolean' then
         for _,y in ipairs(helpersModules[k]) do
            modprobe[#modprobe +1] = ("modprobe "..y.."; ") or nil
         end
      elseif v and type(v)=='table' and v.enable then
         local params = {}
         for x,y in pairs(v) do
            params[#params + 1] = x~='enable' and x.."="..y  or nil
         end
         for x,y in ipairs(helpersModules[k]) do
            modprobe[#modprobe +1] = ("modprobe "..y.." "..(x==1 and table.concat(params," ") or "").."; ") or nil
         end
      end
   end
   return table.concat(modprobe,"\n")
end

local function unloadKernelModules()
   local tmpName = os.tmpname()
   local tmpName2 = os.tmpname()
   local modprobe = insKernelModules()
   local script = ([==[#!/bin/sh
                         {
                            sleep 10;
                            cd /tmp;
                            TMPF=$1;#`mktemp`;
                            R=`uname -r`;
                            {
                               {
                                  ls -t1 /lib/modules/${R}/kernel/net/*/netfilter/ ;
                                  ls -t1 /lib/modules/${R}/kernel/net/netfilter/ ;
                                  ls -t1 /lib/modules/${R}/extra/ipt_* ;
                               } | grep -oe '[^/]*[.]ko$' | grep -oe '^[^.]*';
                            } > $TMPF;
                            MOD='X';
                            i=0;
                            while [ "${MOD}"'X' !=  'X' ]; do
                               MOD=`lsmod | tail -n +2 | grep -E '^[^[:space:]]*[[:space:]]+[^[:space:]]+[[:space:]]+0[[:space:]]+'  | grep -oe '^[^ ]*' | grep -F -f $TMPF ;`
                               if [ "${MOD}"'X' !=  'X' ]; then
                                  rmmod ${MOD};
                                  fi;
                                  i=$(($i+1));
                                  if [ "$i" -gt "7" ]; then
                                     break;
                                     fi
                                     done;
                                    # rm $TMPF;
                               ]==]..modprobe..[==[
                                  } 1>/dev/null 2>/dev/null;]==])
                                  local sFile = io.open(tmpName,"w");
                                  if sFile then
                                     sFile:write(script)
                                     sFile:close()
                                  end
                                  C.restartDaemon(
                                     {"ip", "nf"},
                                     {"/bin/sh", tmpName, tmpName2},
                                     {noRestart = true,
                                      tmpFileList={tmpName, tmpName2},
                                   }
                                  )

end

local function setFilterRulesAux(path, tbl, chain, conf, zeroConf, fixAlias)
  local err = {}
  local function xtest(com)
    com = " -t "..tbl.." "..com
     return os.execute(xtables.iptables[path[1]]..com.." 1>/dev/null  2>/dev/null")
  end
  local function x(com)
    com = xtables.iptables[path[1]].." -t "..tbl.." "..com
    local res,o1,o2 = s.executeEx(com)
    if res ~=0 then
       C.log_debug(com..'\n'..tostring(o2))
       err[#err+1] = '\n'..com..'\n'..(o2:gsub("\nTry[^\n]*\n$", ''))
       return false
    end

    return true
  end

  local pl=tonumber(s.readpopen("cat /proc/sys/kernel/printk | grep -oE '^[0-9]'")) or 7
  os.execute("echo 4 >/proc/sys/kernel/printk")

  if not fixAlias then
     if (chainNameVal(chain) or chain=="UPnP"  or chain=="GRE_KA" ) and (xtest("-n -L "..chain) ~= 0) then
        --create userchain if not created before
        x("-N "..chain)
     end
     x("-F "..chain)
     C.setOption(s.clonePath(path, -1), chain, nil)
     if #err > 0 then
        os.execute("echo "..tostring(pl).." >/proc/sys/kernel/printk")
        return false, err[1]
     end
  end

  if ((not zeroConf) or (conf[0] and ((tbl=="mangle" and chain=="INPUT") or (tbl=="nat" and chain=="PREROUTING") or (tbl=="filter" and chain=="FORWARD"))) or chain=="GRE_KA") and (chain~="UPnP")  then
     local ruleIdx = 0
     for idx = 0, #conf do
        local rule = conf[idx]
        if rule then
           local com
           if fixAlias then
              com = {"-R "..chain..' '..tostring(C.getOption(s.clonePath(path, nil, idx), 'idx') or 0)}
           else
              com = {"-A "..chain}
           end
           if not fixAlias or C.getOption(s.clonePath(path, nil, idx), 'iface') then
              for _, crit in ipairs(criterionPriority) do
                 local val = rule[crit]
                 if crit == "in-interface" or crit == "out-interface" then
                    -- may be interface aliases
                    val = iface_mod.getIfaceName(val)
                    C.setOption(s.clonePath(path, nil, idx), 'iface', true)
                 elseif (crit == "target" or crit == "chain") and (chainNameVal(val) or val=="UPnP" or val=="GRE_KA") and val~="" and val~="NONE"  and (xtest("-n -L "..val) ~= 0) then
                    --precreate userchain if needed for jump to it
                    x("-N "..val)
                    if #err > 0 then
                       return false, err[1]
                    end
                 elseif (crit == "target") and (val=="NETFLOW") then
                    if os.execute("modprobe ipt_NETFLOW")~=0  then
                       return false, 'Can`t load NETFLOW target module'
                    end
                 elseif (crit == "target" or crit == "chain") and (val=="" or val=="NONE") then
                    val = nil
                 elseif (crit == "explicit-matches") and type(val)=="table" then
                    local exp={""}
                    local good=false
                    for mod,parms in pairs(val) do
                       if explicitModulesParams[mod] then
                          exp[#exp + 1] = "-m "..mod
                          if type(parms)=="table" then
                             for k,v in pairs(parms) do
                                local pat = explicitModulesParams[mod][k]
                                if v and pat then
                                   local val, prefix = splitVal(v)
                                   exp[#exp+1] = prefix..pat:format(val)
                                   good = true
                                end
                             end
                          end
                       end
                    end
                    if good then com[#com+1] = table.concat(exp,' ') end
                 end
                 local pat = criterionPatterns[crit]
                 if val and pat then
                    local val, prefix = splitVal(val)
                    com[#com+1] = prefix..pat:format(val)
                 end
              end
              if x(table.concat(com, ' ')) then
                 ruleIdx = ruleIdx + 1
                 if not fixAlias and C.getOption(s.clonePath(path, nil, idx), 'iface') then
                    C.setOption(s.clonePath(path, nil, idx), 'idx', ruleIdx)
                 end
              end
           end
        end
     end
     if conf["default-target"] then
        x("-P "..chain..' '..conf["default-target"])
     end
  end
  if #err > 0 then
    os.execute("echo "..tostring(pl).." >/proc/sys/kernel/printk")
    return false, table.concat(err)
  end

  unloadKernelModules()

  os.execute("echo "..tostring(pl).." >/proc/sys/kernel/printk")
  return true
end

--subitems added automatically
local filterAliasHookKey = {
   ip = {
      nat = {},
      filter = {},
      mangle = {},
   },
   ipv6 = {
      nat = {},
      filter = {},
      mangle = {},
   }
}



local getUserChains =
   function (path, dest)
      local conf = C.getConfByPath(path) or {}
      for k,_ in pairs(conf) do
         if (tostring(k)):match("^[%l]") then
            dest[#dest + 1] = tostring(k)
         end
         end
         table.sort(dest)
      return dest
   end


--this is NOT scheme
--this is array with scheme corrections depends on chains
local iptableSchemeCorrections = {
  PREROUTING = {
    ["out-interface"] = s.nilType,
  },
  INPUT = {
    ["out-interface"] = s.nilType,
  },
  FORWARD = {},
  OUTPUT = {
    ["in-interface"] = s.nilType,
    ["mac-source"] = s.nilType,
  },
  POSTROUTING = {
    ["in-interface"] = s.nilType,
    ["mac-source"] = s.nilType,
  },
  USER = {
     chain =
        function(_, path)
           local targets = getUserChains(s.clonePath(path, -1) or {}, {})
           return not C.getConfByPath(s.clonePath(path, -1, "target")) and
           s.choiceType{
           _H = "nfChain",
           _A = targets,
           _DEP = 1,
        }
     end
  }
}


--this is NOT scheme
--this is array with TARGET and HELP definitions for scheme generation
local iptableTargets = {
  nat = {
    _H = "nfFilterRule",
    PREROUTING = {
      'ACCEPT',
      'CLUSTERIP',
      'CONNMARK',
      'DNAT',
      'DROP',
      'LOG',
      'MARK',
      'NETMAP',
      'NFLOG',
      'RATEEST',
      'REDIRECT',
      'RETURN',
      'TCPMSS',
      'ULOG',
       'NONE',
      --'NFQUEUE',
    },
    POSTROUTING = {
      'ACCEPT',
      'CLUSTERIP',
      'CONNMARK',
      'DROP',
      'LOG',
      'MARK',
      'MASQUERADE',
      'NETMAP',
      'NFLOG',
      'RATEEST',
      'RETURN',
      'SNAT',
      'TCPMSS',
      'ULOG',
       'NONE',
      --'NFQUEUE',
    },
    OUTPUT = {
      'ACCEPT',
      'CLUSTERIP',
      'CONNMARK',
      'DNAT',
      'DROP',
      'LOG',
      'MARK',
      'NETMAP',
      'NFLOG',
      'RATEEST',
      'REDIRECT',
      'RETURN',
      'TCPMSS',
      'ULOG',
       'NONE',
      --'NFQUEUE',
    },
    USER = {
      'ACCEPT',
      'CLUSTERIP',
      'CONNMARK',
      'DNAT',
      'DROP',
      'LOG',
      'MARK',
      'MASQUERADE',
      'NETMAP',
      'NFLOG',
      'RATEEST',
      'REDIRECT',
      'RETURN',
      'SNAT',
      'TCPMSS',
      'ULOG',
       'NONE',
      --'NFQUEUE',
    },
  },
  filter = {
    _H = "nfFilterRule",
    FORWARD = {
      'ACCEPT',
      'CLUSTERIP',
      'CONNMARK',
      'DROP',
      'LOG',
      'MARK',
      'NFLOG',
      'RATEEST',
      'REJECT',
      'RETURN',
      'TCPMSS',
      'ULOG',
       'NETFLOW',
       'NONE',
      --'NFQUEUE',

    },

    INPUT = {
      'ACCEPT',
      'CLUSTERIP',
      'CONNMARK',
      'DROP',
      'LOG',
      'MARK',
      'NFLOG',
      'RATEEST',
      'REJECT',
      'RETURN',
      'TCPMSS',
      'ULOG',
       'NETFLOW',
       'NONE',
      --'NFQUEUE',
    },

    OUTPUT = {
      'ACCEPT',
      'CLUSTERIP',
      'CONNMARK',
      'DROP',
      'LOG',
      'MARK',
      'NFLOG',
      'RATEEST',
      'REJECT',
      'RETURN',
      'TCPMSS',
      'ULOG',
       'NETFLOW',
       'NONE',
      --'NFQUEUE',
    },
    USER = {
      'ACCEPT',
      'CLUSTERIP',
      'CONNMARK',
      'DROP',
      'LOG',
      'MARK',
      'NFLOG',
      'RATEEST',
      'REJECT',
      'RETURN',
      'TCPMSS',
      'ULOG',
       'NETFLOW',
       'NONE',
      --'NFQUEUE',
    },
 },
  mangle = {
    _H = "nfMangleRule",
    PREROUTING = {
      'ACCEPT',
      'CLUSTERIP',
      'CONNMARK',
      'DROP',
      'DSCP',
      'ECN',
      'LOG',
      'MARK',
      'NFLOG',
      'RATEEST',
      'RETURN',
      'TCPMSS',
      'TCPOPTSTRIP',
      'TOS',
      'TPROXY',
      'TTL',
      'ULOG',
       'NONE',
      --'HL',  --only for IPv6
      'NFQUEUE',
    },
    INPUT = {
      'ACCEPT',
      'CLUSTERIP',
      'CONNMARK',
      'DROP',
      'DSCP',
      'ECN',
      'LOG',
      'MARK',
      'NFLOG',
      'RATEEST',
      'RETURN',
      'TCPMSS',
      'TCPOPTSTRIP',
      'TOS',
      'TTL',
      'ULOG',
       'NONE',
      --'HL',  --only for IPv6
      'NFQUEUE',
    },
    FORWARD = {
      'ACCEPT',
      'CLASSIFY',
      'CLUSTERIP',
      'CONNMARK',
      'DROP',
      'DSCP',
      'ECN',
      'LOG',
      'MARK',
      'NFLOG',
      'RATEEST',
      'RETURN',
      'TCPMSS',
      'TCPOPTSTRIP',
      'TOS',
      'TTL',
      'ULOG',
       'NONE',
      --'HL',  --only for IPv6
      'NFQUEUE',
    },
    OUTPUT = {
      'ACCEPT',
      'CLASSIFY',
      'CLUSTERIP',
      'CONNMARK',
      'DROP',
      'DSCP',
      'ECN',
      'LOG',
      'MARK',
      'NFLOG',
      'RATEEST',
      'RETURN',
      'TCPMSS',
      'TCPOPTSTRIP',
      'TTL',
      'ULOG',
       'NONE',
      --'HL',  --only for IPv6
      'NFQUEUE',
    },
    POSTROUTING = {
      'ACCEPT',
      'CLASSIFY',
      'CLUSTERIP',
      'CONNMARK',
      'DROP',
      'DSCP',
      'ECN',
      'LOG',
      'MARK',
      'NFLOG',
      'RATEEST',
      'RETURN',
      'TCPMSS',
      'TCPOPTSTRIP',
      'TOS',
      'TTL',
      'ULOG',
       'NONE',
      --'HL',  --only for IPv6
      'NFQUEUE',
    },
    USER = {
      'ACCEPT',
      'CLASSIFY',
      'CLUSTERIP',
      'CONNMARK',
      'DROP',
      'DSCP',
      'ECN',
      'LOG',
      'MARK',
      'NFLOG',
      'RATEEST',
      'RETURN',
      'TCPMSS',
      'TCPOPTSTRIP',
      'TOS',
      'TPROXY',
      'TTL',
      'ULOG',
       'NONE',
      --'HL',  --only for IPv6
      'NFQUEUE',
    },
  },
}

local iptableTargetOptions = {
  CLASSIFY = {
    ["set-class"] = s.stringType{
      _H = "nfClassifySetclass",
      _P = -90,
    },
  },
  CLUSTERIP = {
    ["new"] = s.booleanType{
      _H = "nfClusteripNew",
      _P = -90,
    },
    ["hashmode"] = s.choiceType{
      _H = "nfClusteripHashmode",
      _A = {
        'sourceip',
        'sourceip-sourceport',
        'sourceip-sourceport-destport',
      },
      _P = -90,
    },
    ["clustermac"] = s.stringType{
      _V = s.macaddressVal,
      _H = "nfClusteripClustermac",
      _P = -90,
    },
    ["total-nodes"] = s.integerType{
      _H = "nfClusteripTotalnodes",
      _P = -90,
    },
    ["local-node"] = s.integerType{
      _H = "nfClusteripLocalnode",
      _P = -90,
    },
    ["hash-init"] = s.integerType{
      _H = "nfClusteripHashinit",
      _P = -90,
    },
  },
  CONNMARK = {
    ["connmark-params"] = s.stringType{
      _H = "nfConnmarkParams",
      _P = -90,
    },
  },
  DNAT = {
    ["to-destination"] = s.stringType{
      _H = "nfDnatTodestination",
      _P = -90,
    },
    ["random"] = s.booleanType{
      _H = "nfDnatRandom",
      _P = -90,
    },
  },
  -- 'DROP'
  DSCP   = {
    ["set-dscp"] = s.integerType{
      _H = "nfDscpSetdscp",
      _P = -90,
    },
    ["set-dscp-class"] = s.stringType{
      _H = "nfDscpSetdscpclass",
      _P = -90,
    },
  },
  ECN = {
    ["ecn-tcp-remove"] = s.booleanType{
      _H = "nfEcnEcntcpremove",
      _P = -90,
    },
  },
  --'HL',  --only for IPv6
  LOG = {
    ["log-level"] = s.integerType{
      _H = "nfLogLoglevel",
      _P = -90,
    },
    ["log-prefix"] = s.stringType{
      _H = "nfLogLogprefix",
      _P = -90,
    },
    ["log-tcp-sequence"] = s.booleanType{
      _H = "nfLogLogtcpsequence",
      _P = -90,
    },
    ["log-tcp-options"] = s.booleanType{
      _H = "nfLogLogtcpoptions",
      _P = -90,
    },
    ["log-ip-options"] = s.booleanType{
      _H = "nfLogLogipoptions",
      _P = -90,
    },
    ["log-uid"] = s.booleanType{
      _H = "nfLogLoguid",
      _P = -90,
    },
  },
  MARK = {
    ["set-mark"] = s.stringType{
      _H = "nfMarkSetmark",
      _P = -90,
    },
    ["set-xmark"] = s.stringType{
      _H = "nfMarkSetxmark",
      _P = -90,
    },
  },
  MASQUERADE = {
    ["to-ports"] = s.stringType{
      _H = "nfMasqueradeToports",
      _P = -90,
    },
    ["random"] = s.booleanType{
      _H = "nfMasqueradeRandom",
      _P = -90,
    },
  },
  NETMAP = {
    ["to"] = s.stringType{
      _H = "nfNetmapTo",
      _P = -90,
    },
  },
  --'NFLOG',
  NFQUEUE = {
     ["queue-num"] = s.integerType{
       _H = "nfNfqueueNum",
       _A = {0, 65535},
       _P = -90,
     },
  },
  RATEEST = {
    ["rateest-name"] = s.stringType{
      _H = "nfRateestName",
      _P = -90,
    },
    ["rateest-interval"] = s.stringType{
      _H = "nfRateestInterval",
      _P = -90,
    },
    ["rateest-ewmalog"] = s.integerType{
      _H = "nfRateestEwmalog",
      _P = -90,
    },
  },
  REDIRECT = {
    ["to-ports"] = s.stringType{
      _H = "nfRedirectToports",
      _P = -90,
    },
    ["random"] = s.booleanType{
      _H = "nfRedirectRandom",
      _P = -90,
    },
  },
  REJECT = {
    ["reject-with"] = s.choiceType{
      _H = "nfRejectwith",
      _A = {
        'icmp-net-unreachable',
        'icmp-host-unreachable',
        'icmp-port-unreachable',
        'icmp-proto-unreachable',
        'icmp-net-prohibited',
        'icmp-host-prohibited',
        'icmp-admin-prohibited',
      },
      _P = -90,
    },
  },
  --'RETURN',
  SNAT = {
    ["to-source"] = s.stringType{
      _H = "nfSnatTosource",
      _P = -90,
    },
    ["random"] = s.booleanType{
      _H = "nfSnatRandom",
      _P = -90,
    },
  },
  TCPMSS = {
    ["set-mss"] = s.integerType{
      _H = "nfTcpmssSetmss",
      _P = -90,
    },
    ["clamp-mss-to-pmtu"] = s.booleanType{
      _H = "nfTcpmssClamptopmtu",
      _P = -90,
    },
  },
  TCPOPTSTRIP = {
    ['strip-options'] = s.stringType{
      _H = "nfTcpoptstripOptions",
      _P = -90,
    },
  },
  TOS = {
    ['set-tos'] = s.stringType{
      _H = "nfTosSet",
      _P = -90,
    },
  },
  --'TPROXY',
  TTL = {
    ['ttl-set'] = s.stringType{
      _H = "nfTtlSet",
      _P = -90,
    },
    ['ttl-dec'] = s.stringType{
      _H = "nfTtlDec",
      _P = -90,
    },
    ['ttl-inc'] = s.stringType{
      _H = "nfTtlInc",
      _P = -90,
    },
  },
  ULOG = {
    ["ulog-nlgroup"] = s.integerType{
      _H = "nfUlogNlgroup",
      _P = -90,
      _A = {1,32}
    },
    ["ulog-prefix"] = s.stringType{
      _H = "nfUlogPrefix",
      _P = -90,
    },
    ["ulog-cprange"] = s.integerType{
      _H = "nfUlogCprange",
      _P = -90,
    },
    ["ulog-qthreshold"] = s.integerType{
      _H = "nfUlogQthreshold",
      _P = -90,
    },
  },
  EMPTY = {
  }
}

local iptableExplicitMatches = s.newScheme{
  addrtype = s.newScheme{
    _H = "nfCriterionAddrtype",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  ah = s.newScheme{
    _H = "nfCriterionAh",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  comment = s.newScheme{
    _H = "nfCriterionComment",
    ["comment"] =  s.stringType{_H = "nfCriterionComment"},},
  connbytes = s.newScheme{
    _H = "nfCriterionConnbytes",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  connlimit = s.newScheme{
    _H = "nfCriterionConnlimit",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  connmark = s.newScheme{
    _H = "nfCriterionConnmark",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  conntrack = s.newScheme{
    _H = "nfCriterionConntrack",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  dscp = s.newScheme{
    _H = "nfCriterionDscp",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  ecn = s.newScheme{
    _H = "nfCriterionEcn",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  esp = s.newScheme{
    _H = "nfCriterionEsp",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  hashlimit = s.newScheme{
    _H = "nfCriterionHashlimit",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  helper = s.newScheme{
    _H = "nfCriterionHelper",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  iprange = s.newScheme{
    _H = "nfCriterionIprange",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  length = s.newScheme{
    _H = "nfCriterionLength",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  limit = s.newScheme{
    _H = "nfCriterionLimit",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  mac = s.newScheme{
    _H = "nfCriterionMac",
    ["mac-source"] = s.stringType{
      _V = genCheckValOrInverse(s.macaddressVal),
      _H = "nfCriterionMacSource",
    },
  },
  mark = s.newScheme{
    _H = "nfCriterionMark",
    ["mark"] = s.stringType{ _H = "nfCriterionMark", },
  },
  multiport = s.newScheme{
    _H = "nfCriterionMultiport",
    ["source-ports"] = s.stringType{ _H = "nfCriterionSports", },
    ["destination-ports"] = s.stringType{ _H = "nfCriterionDports", },
  },
  --owner = s.newScheme{},
  physdev = s.newScheme{
    _H = "nfCriterionPhysdev",
    ["physdev-in"] = s.stringType{ _H = "nfCriterionPhysdevIn", },
    ["physdev-out"] = s.stringType{ _H = "nfCriterionPhysdevOut", },
    ["physdev-is-in"] = s.booleanType{ _H = "nfCriterionPhysdevIsin", },
    ["physdev-is-out"] = s.booleanType{ _H = "nfCriterionPhysdevIsout", },
    ["physdev-is-bridged"] = s.booleanType{ _H = "nfCriterionPhysdevIsbridged", },
  },
  pkttype = s.newScheme{
    _H = "nfCriterionPkttype",
    ["pkt-type"] = s.choiceType{
      _H = "nfCriterionPkttype",
      _A = withInverted({"unicast", "broadcast", "multicast"}),
    }
  },
  policy = s.newScheme{
    _H = "nfCriterionPolicy",
    ["dir"] = s.choiceType{
      _H = "nfCriterionPolicyDir",
      _A = {"in", "out"}
    },
    ["pol"] = s.choiceType{
      _H = "nfCriterionPolicyPol",
      _A = {"none", "ipsec"}
    },
    ["strict"] = s.booleanType{_H = "nfCriterionPolicyStrict",},
    ["reqid"] = s.stringType{_H = "nfCriterionPolicyReqid",},
    ["spi"] = s.stringType{_H = "nfCriterionPolicySpi",},
    ["proto"] = s.choiceType{
      _H = "nfCriterionPolicyProto",
      _A = withInverted{"ah", "esp", "ipcomp"},
    },
    ["mode"] = s.choiceType{
      _H = "nfCriterionPolicyMode",
      _A = withInverted{"tunnel", "transport"},
    },
    ["tunnel-src"] = s.stringType{_H = "nfCriterionPolicyTunnelsrc",},
    ["tunnel-dst"] = s.stringType{_H = "nfCriterionPolicyTunneldst",},
    ["next"] = s.booleanType{_H = "nfCriterionPolicyNext",},
  },
  ["quota"] = s.newScheme{
    _H = "nfCriterionQuota",
    ["quota"] = s.integerType{_H = "nfCriterionQuota",}
  },
  rateest = s.newScheme{
    _H = "nfCriterionRateest",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},
  },
  realm = s.newScheme{
    _H = "nfCriterionRealm",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  recent = s.newScheme{
    _H = "nfCriterionRecent",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  sctp = s.newScheme{
    _H = "nfCriterionSctp",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  --[[   set = s.newScheme{
      _H = "nfCriterionSet",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
--]]
  state = s.newScheme{
    _H = "nfCriterionState",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  statistic = s.newScheme{
    _H = "nfCriterionStatistic",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  string = s.newScheme{
    _H = "nfCriterionString",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  tcpmss = s.newScheme{
    _H = "nfCriterionTcpmss",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  time = s.newScheme{
    _H = "nfCriterionTime",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  tos = s.newScheme{
    _H = "nfCriterionTos",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  ttl = s.newScheme{
    _H = "nfCriterionTtl",
    ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  u32 = s.newScheme{
    _H = "nfCriterionU32",
    ["value"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  --[[ unclean = s.booleanType{
      _H = "nfCriterionUnclean",
      },
--]]
}

local function setFilterRules(path, sch, conf, prevConf)
   path=s.deepcopy(path)
   s.confValidate(path, sch, conf, true)
   local zeroConf = (conf==nil or s.confCmp(conf, {})) and true or false
   local dconf = s.addDefault(path, sch, conf, true)
   local sdconf = nsg.serialize(dconf)

   if  sdconf~=C.getOption(path, 'defConf') or not s.confCmp(conf, prevConf) then
      C.setOption(path, 'defConf', sdconf)
      local tbl = path[#path-1]
      local chain = path[#path]
      local ipv = path[#path-2]
      filterAliasHookKey[ipv][tbl][chain] = filterAliasHookKey[ipv][tbl][chain]  or {}
      iface_mod.setAliasHook(filterAliasHookKey[ipv][tbl][chain],
                             dconf[1] and
                                function()
                                   return  setFilterRulesAux(path ,tbl, chain, dconf, nil, true)
                                end or nil
                          )
      return setFilterRulesAux(path, tbl, chain, dconf, zeroConf)
   end
   return true
end

local portRange = s.newScheme{
  _T = 'string',
  _V = function(val)
    if type(val) == 'number' and val >= 0 and val <= 0xffff and val%1==0 then
      return tostring(val)
    end
    if type(val) == 'string' then
      local negative, p1, p2 = val:match"^(!?)(%d%d?%d?%d?%d?)(:?%d?%d?%d?%d?%d?)$"
      if (
        p1 and tonumber(p1) <= 0xffff  and
        (p2=='' or p2:sub(1,1) == ':' and tonumber(p2:sub(2)) <= 0xffff)
      ) then
        return val
      end
    end
    return nil, "port range in format  port[:port]"
  end
}

local criterionScheme = s.newScheme{
   protocol = s.newScheme{
      _H = "nfCriterionProtocol",
      _T = "choice",
      _A = getProtocols(),
      _V = s.choiceVal,
      _P = 60,
      _DEP = 1,
   },
   ["protocol-num"] =
      function(_, path)
         return not C.getConfOrDefault(s.clonePath(path, -1, "protocol")) and
         s.integerType{
         _H = "nfCriterionProtocolNum",
         _A = {0, 0xffff},
         _P = 50,
      }
   end,
  ["tcp-flags"] =
      function(_, path)
         local proto = C.getConfOrDefault(s.clonePath(path, -1, "protocol"))
         return (proto == 'tcp' ) and
         s.stringType{
         _H = "nfCriterionTcpflags",
         _P = 7,
      }
   end,
   ["in-interface"] = s.stringType{
      _H = "nfCriterionInInterface",
      _P = 40,
   },
   ["out-interface"] = s.stringType{
      _H = "nfCriterionOutInterface",
      _P = 30,
   },
   source = s.stringType{
      _H = "nfCriterionSource",
      _P = 20
   },
   destination = s.stringType{
      _H = "nfCriterionDestination",
      _P = 10,
   },
   ["source-port"] =
      function(_, path)
         local proto = C.getConfOrDefault(s.clonePath(path, -1, "protocol"))
         return (proto == 'tcp' or proto == 'udp') and
         portRange{
         _H = "nfCriterionSport",
         _P = 8,
      }
   end,
   ["destination-port"] =
      function(_, path)
         local proto = C.getConfOrDefault(s.clonePath(path, -1, "protocol"))
         return (proto == 'tcp' or proto == 'udp') and
         portRange{
         _H = "nfCriterionDport",
         _P = 7,
      }
   end,
   ["icmp-type"] =
      function(_, path)
         local proto = C.getConfOrDefault(s.clonePath(path, -1, "protocol"))
         return (path[1]=='ip' and  (proto == 'icmp' or proto == '!icmp')) and
         s.newScheme{
         _H = "nfCriterionIcmpType",
         _T = "choice",
         _A = getIcmpType({xtables.icmpproto[proto]}),
         _V = s.choiceVal,
         _P = 7,
      }
   end,
   ["icmpv6-type"] =
      function(_, path)
         local proto = C.getConfOrDefault(s.clonePath(path, -1, "protocol"))
         return (path[1]=='ip' and  (proto == 'ipv6-icmp' or proto == '!ipv6-icmp')) and
         s.newScheme{
         _H = "nfCriterionIcmpType",
         _T = "choice",
         _A = getIcmpType({xtables.icmpproto[proto]}),
         _V = s.choiceVal,
         _P = 7,
      }
   end,
   ["explicit-matches"] = iptableExplicitMatches{
      _H = "nfExplicitMatches",
      _P = 6,
   },
   ["extra-options"] = s.stringType{
      _H = "nfCriterionExtra",
      _P = 1,
   },
   ["extra-target"] = s.stringType{
      _H = "nfExtraTarget",
      _P = -100,
   },
}

local criterionParms  =
function(_, path)
   local targets = iptableTargets[getTable(path)][getChain(path)]
  or iptableTargets[getTable(path)]['USER'] or {}
  targets = s.deepcopy(targets)
  if (getTable(path)=="nat" or getTable(path)=="filter") and upnpEnabled(path) then
     targets[#targets+1] = 'UPnP'
  end
  if (getTable(path)=="mangle") then
     targets[#targets+1] = 'GRE_KA'
  end
  targets = getUserChains(s.clonePath(path, -2) or {}, targets)
 
  local sch1 = criterionScheme({
      _H = iptableTargets[getTable(path)]._H,
      target = function(_, path)
                  return not C.getConfByPath(s.clonePath(path, -1, "chain")) and
                  s.choiceType{
                     _H = "nfTarget",
                     _A = targets,
                     _DEP = 1,
                  }
               end,
  })

  local target = C.getConfByPath(s.clonePath(path, nil, "target")) or "EMPTY"
  local sch2 = s.newScheme(iptableTargetOptions[target] or {})
  local sch3 = s.newScheme(iptableSchemeCorrections[path[#path-1]] or iptableSchemeCorrections['USER'] or {})

  return sch1(sch2)(sch3)(_, path)
end

upnpEnabled =  function()
                  local upnp = C.getConfByPath({"services", "upnp"}) or {}
                  --  for k,v in pairs(upnp) do
                  --                      if v["adm-state"] == "up" then
                  if upnp["adm-state"] == "up" then
                     return true
                  end
                  --  end
                  return nil
               end

upnpRule = function(_, path)
              local targets = getUserChains(s.clonePath(path, -1) or {}, {"UPnP", "NONE"})
              return upnpEnabled() and s.newScheme{
                 _H = "nfUpnpRule",
                 target = s.choiceType{
                    _A = targets,
                    _D = 'UPnP'
                 }
              } or nil
           end

upnpChain = function(_, path)
               return upnpEnabled() and s.newScheme{
                  _Z = setFilterRules,
                  _P = 7,
                  _H = "nfUpnpChain",
                  show = s.newScheme{
                     _H = "nfFilterUPnPShow",
                     _X = function(path)
                             return('\n'..s.readpopen(xtables.iptables[path[1]].."  -nvL UPnP  -t "..path[#path -2].." 2>&1"))
                          end,
                     _R = 'user',
                  },
                --[[  ["default-target"] = s.choiceType{
                     _A = {'ACCEPT'},
                     _D = 'ACCEPT'
                  }
               --]]
               } or nil
end

grekaEnabled =  function()
                   local gre = C.getConfByPath({"tunnel", "gre"}) or {}
                   local greS = C.getSchemeByPath({"tunnel", "gre"}) or s.newScheme{}
                   gre = s.addDefault({"tunnel", "gre"}, greS, gre, true)
                   for k,v in pairs(gre) do
                      if tostring(k):match("^gre[0-9]+$") then
                         if v["adm-state"] == "up" and v.encapsulation=="ip" and v.keepalive=="yes" and s.ipaddressVal(v.source) then
                            return true
                         end
                      end
                   end
                   return nil
                end

grekaRule = function(_, path)
              local targets = getUserChains(s.clonePath(path, -1) or {}, {"GRE_KA", "NONE"})
              return grekaEnabled() and s.newScheme{
                 _H = "nfGrekaRule",
                 protocol = s.choiceType{
                       _A = {"gre"},
                       _D = "gre"
                    },
                 target = s.choiceType{
                    _A = targets,
                    _D = 'GRE_KA',
                 }
              } or nil
           end

grekaChain = function(_, path)
               return grekaEnabled() and s.newScheme{
                  _Z = setFilterRules,
                  _P = 7,
                  _H = "nfGrekaChain",
                  show = s.newScheme{
                     _H = "nfFilterGrekaShow",
                     _X = function(path)
                             return('\n'..s.readpopen(xtables.iptables[path[1]].."  -nvL GRE_KA  -t "..path[#path -2].." 2>&1"))
                          end,
                     _R = 'user',
                  },
                  [1] = s.newScheme{
                     _H = "nfGreka1Rule",
                     ["explicit-matches"] = s.newScheme{
                        _H = "nfExplicitMatches",
                        u32 = s.newScheme{
                           _H = "nfCriterionU32",
                           value = s.stringType{
                              _H = "nfCriterionU32",
                              _D = "0>>22&0x3C@0&0xD807FFFF=0x0",--IpHeadSize@GreProto
                           }
                        }
                     },
                     target = s.choiceType{
                        _A = {"NFQUEUE"},
                        _D = "NFQUEUE"
                     },
                     ["queue-num"] = s.choiceType{
                        _A = {"50009"},
                        _D = "50009"
                     }
                  },
                  [2] = s.newScheme{
                     _H = "nfGreka2Rule",
                     ["explicit-matches"] = s.newScheme{
                        _H = "nfExplicitMatches",
                        u32 = s.newScheme{
                           _H = "nfCriterionU32",
                           value = s.stringType{
                              _H = "nfCriterionU32",
                              _D = "0>>22&0x3C@0&0xF807FFFF=0x00000800 && 0>>22&0x3C@10&0xFF=47 && 0>>22&0x3C@24&0xF807FFFF=0x0",
                              --IpHeadSize@GreProto
                           }
                        }
                     },
                     target = s.choiceType{
                        _A = {"NFQUEUE"},
                        _D = "NFQUEUE"
                     },
                     ["queue-num"] = s.choiceType{
                        _A = {"50009"},
                        _D = "50009"
                     }
                  },
                  [3] = s.newScheme{
                     _H = "nfGreka2Rule",
                     ["explicit-matches"] = s.newScheme{
                        _H = "nfExplicitMatches",
                        u32 = s.newScheme{
                           _H = "nfCriterionU32",
                           value = s.stringType{
                              _H = "nfCriterionU32",
                              _D ="0>>22&0x3C@0&0xF807FFFF=0x20000800 && 0>>22&0x3C@14&0xFF=47 && 0>>22&0x3C@28&0xF807FFFF=0x20000000",
                              --IpHeadSize@GreProto
                           }
                        }
                     },
                     target = s.choiceType{
                        _A = {"NFQUEUE"},
                        _D = "NFQUEUE"
                     },
                     ["queue-num"] = s.choiceType{
                        _A = {"50009"},
                        _D = "50009"
                     }
                  },
               } or nil
            end


local chainScheme = s.newScheme{
  _Z = setFilterRules,
  _LIST = criterionParms,
  ["default-target"] = s.newScheme{
    _H = "nfDefaultTarget",
    _T = "choice",
    _A = {
      'ACCEPT',
      'DROP',
    },
    _D = "ACCEPT",
    _V = s.choiceVal,
  },
}

applyNFNetflow =
   function(path, sch, conf, prevConf)
      if os.execute( "lsmod | grep -oe 'ipt_NETFLOW' 2>/dev/null 1>/dev/null" )==0 then
         conf = C.getConfOrDefault({"ip","filter","NETFLOW"})
         if conf then
            local ctl_r = "sysctl -n net.netflow."
            local ctl_w = "sysctl -w net.netflow."
            local test_and_exec =
               function(name, val)
                  val = val or conf[name]
                  if val and s.readpopen(ctl_r..name:gsub("-","_")):gsub("[^0-9.:/-]","") ~= val then
                     os.execute(ctl_w..name:gsub("-","_").."="..val)
                  end
               end
            test_and_exec("active-timeout")
            test_and_exec("inactive-timeout")
            test_and_exec("aggregation")
            local d
            for k, v in ipairs(conf.destination or {}) do
               if v and type(v)=='string' and #v>0 then
                  d = (d and d.."," or "")..v
               end
            end
            test_and_exec("destination", d)
         end
      end
      return true
   end

local netflowScheme = s.newScheme{
   _Z = applyNFNetflow,
   _H = "nfFilterNetflow",
   _P = 7,
   destination = s.newScheme{
      _H = "nfFilterNetflowDestination",
      [1] = s.ipAddressAndPortType{
         _H = "nfFilterNetflowDestinationN",
         _D = "127.0.0.1:2055",
      },
      _LIST = s.ipAddressAndPortType{
         _H = "nfFilterNetflowDestinationN",
      },
   },
   ["inactive-timeout"] = s.integerType{
      _H = "nfFilterNetflowInactiveTimeout",
      _D = 15,
   },
   ["active-timeout"] = s.integerType{
      _H = "nfFilterNetflowActiveTimeout",
      _D = 1800,
   },
   aggregation = s.stringType{
      _H = "nfFilterNetflowActiveAggregation",
      _A = {700}
   },
   show = s.newScheme{
      _H = "nfFilterNetflowShow",
      _X = function()
              return(s.readpopen("cat /proc/net/stat/ipt_netflow 2>/dev/null"))
           end,
   },

}

local LO_ADDRESS = "127.0.0.1"
local LO_ADDRESS6 = "::1"
local loAddrScheme =
   function (_,path)
      local lo = path[1]=='ip' and LO_ADDRESS or LO_ADDRESS6
      return s.stringType{
         _H = "nfCriterionSource",
         _V = function(val)
                 return val == lo and lo or
                    nil, lo
              end,
         _D = lo,
      }
   end
local onlyAcceptTargetScheme = s.newScheme{
  _H = "nfTarget",
  _T = "choice",
  _A = {
    'ACCEPT',
     'NONE' -- this allow user to deactivate this rule if needed
  },
  _V = s.choiceVal,
  _D = 'ACCEPT',
}
local filterInput0 = s.newScheme{
  _H = "nfFilterInputRule0",
  source = loAddrScheme,
  target = onlyAcceptTargetScheme,
}
local filterOutput0 = s.newScheme{
  _H = "nfFilterOutputRule0",
  destination = loAddrScheme,
  target = onlyAcceptTargetScheme,
}

filterScheme = s.newScheme{
   _H = "nfFilter",
   _Z = function() return true end,
   INPUT = chainScheme{
      _H = "nfFilterInput",
      [0] = filterInput0,
      _P = 30,
   },
   FORWARD = chainScheme{
      _H = "nfFilterForward",
      [0] = upnpRule,
      _P = 20,
   },
   OUTPUT =  chainScheme{
      _H = "nfFilterOutput",
      [0] = filterOutput0,
      _P = 10,
   },
   _NAME_TYPE = chainNameVal,
   _NAME_H = "nfUserChainPrompt",
   _NAME = s.newScheme{
      _H = "nfUserChain",
      _P = 5,
      _DEP = 1,
      _Z = setFilterRules,
      _LIST = criterionParms,
   },
   UPnP = upnpChain,
   NETFLOW = netflowScheme,
   SHOW = s.newScheme{
      _P = 6,
      _H = "nfFilterShow",
      _X = function(path)
              return('\n'..s.readpopen(xtables.iptables[path[1]].."  -nvL -t filter 2>&1"))
           end,
      _R = 'user',
   }
}

--------------------------------------------------------------------------------
-- NAT

natScheme = s.newScheme{
  _H = "nfNat",
  _Z = function() return true end,

  PREROUTING =  chainScheme{
     _H = "nfNatPrerouting",
     [0] =  upnpRule,
     _P = 30,
  },
  POSTROUTING =  chainScheme{
    _H = "nfNatPostrouting",
    _P = 20,
  },
  OUTPUT =  chainScheme{
    _H = "nfNatOutput",
    _P = 10,
  },
  _NAME_TYPE = chainNameVal,
  _NAME_H = "nfUserChainPrompt",
  _NAME = s.newScheme{
    _H = "nfUserChain",
    _P = 5,
    _DEP = 1,
    _Z = setFilterRules,
    _LIST = criterionParms
  },
  UPnP = upnpChain,
  SHOW = s.newScheme{
    _H = "nfNatShow",
    _X = function(path)
            return('\n'..s.readpopen(xtables.iptables[path[1]].."  -nvL -t nat 2>&1"))
    end,
    _R = 'user',
  },
  HELPERS = function(_, path)
               return path[1] == "ip" and
                s.newScheme{
                  _H = "nfNatHelpers",
                  DCCP = s.booleanType{_H = "nfNatHelpersDCCP", },
                  SCTP = s.booleanType{_H = "nfNatHelpersSCTP", },
                  ["UDP-Lite"] = s.booleanType{_H = "nfNatHelpersUdplite", },
                  Amanda = s.newScheme{
                     _H = "nfNatHelpersAmanda",
                     enable = s.booleanType{_H =  "nfNatHelpersAmandaEnable", _P =100},
                     master_timeout = s.integerType{
                        _H = "nfNatHelpersAmandaMastertimeout",
                        _A = {1,2^32 -1},
                        _D = 300,
                     },
                  },
                  FTP = s.newScheme{
                     _H = "nfNatHelpersFtp",
                     enable = s.booleanType{_H = "nfNatHelpersFtpEnable", _P =100},
                     ports = s.stringType{_H = "nfNatHelpersFtpPorts", _D= "21",},
                  },
                  ["H.323"] = s.newScheme{
                     _H = "nfNatHelpersH323",
                     enable = s.booleanType{_H = "nfNatHelpersH323Enable", _P =100},
                     default_rrq_ttl = s.integerType{
                        _H = "nfNatHelpersH323Defaultrrqttl",
                        _A  = {0,2^32-1},
                        _D = 300,
                     },
                     gkrouted_only =  s.choiceType{
                        _H = "nfNatHelpersH323Gkrouted_only",
                        _A  = {1,0},
                        _D = 1,
                     },
                     callforward_filter =  s.choiceType{
                        _H = "nfNatHelpersH323Callforwardfilter",
                        _A  = {1,0},
                        _D = 1,
                     },
                  },

                  IRC = s.newScheme{
                     _H = "nfNatHelpersIrc",
                     enable = s.booleanType{_H = "nfNatHelpersIrcEnable", _P =100},
                     ports = s.stringType{_H = "nfNatHelpersIrcPorts", _D= "6667",},
                     max_dcc_channels = s.integerType{
                        _H = "nfNatHelpersIrcMaxdccchannels",
                        _A = {1,2^32-1},
                        _D= 8,
                     },
                     dcc_timeout = s.integerType{
                        _H = "nfNatHelpersIrcDcctimeout",
                        _A = {1,2^32-1},
                        _D= 300,
                     },
                  },
                  NetBIOS = s.newScheme{
                     _H = "nfNatHelpersNetbios",
                     enable = s.booleanType{_H = "nfNatHelpersNetbiosEnable", _P =100},
                     timeout = s.integerType{
                        _H = "nfNatHelpersNetbiosTimeout",
                        _A = {1,2^32-1},
                        _D= 3,
                     },
                  },
                  PPTP = s.booleanType{_H = "nfNatHelpersPptp", },
                  SANE = s.newScheme{
                     _H = "nfNatHelpersSane",
                     enable = s.booleanType{_H = "nfNatHelpersSaneEnable", _P =100},
                     ports = s.stringType{_H = "nfNatHelpersSanePorts", _D= "6566",},
                  },
                  SIP = s.newScheme{
                     _H = "nfNatHelpersSip",
                     enable = s.booleanType{_H = "nfNatHelpersSipEnable", _P =100},
                     ports = s.stringType{_H = "nfNatHelpersSipPorts", _D= "5060",},
                     sip_timeout = s.integerType{
                        _H = "nfNatHelpersSipTimeout",
                        _A = {1,2^32-1},
                        _D= 3600,
                     },
                     sip_direct_signalling =  s.choiceType{
                        _H = "nfNatHelpersSipSipdirectsignalling",
                        _A  = {1,0},
                        _D = 1,
                     },
                     sip_direct_media =  s.choiceType{
                        _H = "nfNatHelpersSipSipdirectmedia",
                        _A  = {1,0},
                        _D = 1,
                     },
                  },
                  TFTP = s.newScheme{
                     _H = "nfNatHelpersTftp",
                     enable = s.booleanType{_H = "nfNatHelpersTftpEnable", _P =100},
                     ports = s.stringType{_H = "nfNatHelpersTftpPorts", _D= "69",},
                  },
                  SNMP = s.booleanType{_H = "nfNatHelpersSNMP", },
               }
            end ,
}
--------------------------------------------------------------------------------
-- MANGLE

mangleScheme = s.newScheme{
  _H = "nfMangle",
  _Z = function() return true end,
  PREROUTING = chainScheme{
    _H = "nfManglePrerouting",
    _P = 30,
  },
  INPUT =  chainScheme{
    _H = "nfMangleInput",
    _P = 25,
     [0] = grekaRule,
  },
  FORWARD =  chainScheme{
    _H = "nfMangleForward",
    _P = 20,
  },
  OUTPUT =  chainScheme{
    _H = "nfMangleOutput",
    _P = 15,
  },
  POSTROUTING =  chainScheme{
    _H = "nfManglePostrouting",
    _P = 10,
  },
  GRE_KA = grekaChain,
  _NAME_TYPE = chainNameVal,
  _NAME_H = "nfUserChainPrompt",
  _NAME = s.newScheme{
    _H = "nfUserChain",
    _P = 5,
    _DEP = 1,
    _Z = setFilterRules,
    _LIST = criterionParms,
  },
  SHOW = s.newScheme{
    _H = "nfMangleShow",
    _X = function(path)
            return('\n'..s.readpopen(xtables.iptables[path[1]].."  -nvL -t mangle 2>&1"))
    end,
    _R = 'user',
  },
}
