module(..., package.seeall)
local s = require'scheme'
local dhcp = require'nsgdhcp'
local dyndns = require'nsgdyndns'
local netping = require'nsgnetping'
local daemons = require'nsgdaemons'
local snmp = require'nsgsnmp'
local upnp = require'nsgupnpconf'
local event = require'nsgeventconf'
local ixia = require'nsgixiaconf'
local tp = require'nsgtpconf'
local netflow = require'nsgnetflow'

local sshdConfigFileName = "/etc/ssh/sshd_config"
local sshdConfigFileContent = [[
Protocol 2
UsePAM yes
Subsystem	sftp	/usr/libexec/sftp-server
]]

local function restartSSHDaemon(path, sch, conf, prevConf)
  if not s.confCmp(conf, prevConf) then
    conf = conf or {}
    if conf.enable then
      -- create config.file
      local f, err = io.open(sshdConfigFileName, 'w')
      if f then
        f:write(sshdConfigFileContent)
        f:close()
      else
        C.log_error("Can't create sshd config file: %s" % err)
      end

      local sshconf = s.addDefault(path, sch, conf, true) --clone conf with defaults
      local com = {
        "/sbin/sshd",
        "-D",
        "-e",
        "-p", tostring(sshconf.port or 22),
        sshconf["debug-level"] > 0 and '-'..("d"):rep(sshconf["debug-level"])
        or nil -- must be last element in this table
      }
      C.restartDaemon(path, com, {atExit = C.increasingSleep})
    else
      C.restartDaemon(path)
    end
  end
  return true
end

local function restartHttpDaemon(path, sch, conf, prevConf, first, sesNum)
  local si = sesNum and C.getSessionInfo(sesNum)
  local needCert = 0 == os.execute('cd /etc/stunnel/;[ ! -s https_auto.key ] || [ ! -s https_auto.cert ]')
  local httpconf = s.addDefault(path, sch, conf, true) --clone conf with defaults
  if (not s.confCmp(conf, prevConf)) or
  ((needCert and httpconf.enable and httpconf.https=="enable")) then
    if si == nil or (si and si.mode ~= "html") or (si and si.EXTRA["http-apply-warn"]) then
      if httpconf.enable then
        if httpconf.https=="enable" then
          local certGenScript= [[  cd /etc/stunnel/;
                     {
                        if [ ! -s https_auto.key ] || [ ! -s https_auto.cert ];
                        then
                              /bin/openssl req -new -x509 -days 3650 -nodes -out https_auto.cert -keyout https_auto.key -config /etc/ssl/openssl.cnf -subj /C=RU/ST=Russia/L=`hostname`/O=`hostname`/OU=IT/CN=`hostname`/ ;chmod 600  https_auto.key;
                        else
                           echo 'cert and key for https already exist, skip generation'
                           fi; } 2>/dev/null;
                        exit 0;
                  ]]
          C.restartDaemon(
            {"services","https"},
            {"/bin/sh", "-c", certGenScript},
            {noRestart = true}
          )
        end

        --[[
            local com = {
               "/usr/sbin/httpd",
               "-f",
               "-h"..httpconf.home or "-h/usr/share/www/",
               httpconf.port and '-p'..tonumber(httpconf.port) or nil
            }
            --]]
        local com = {
           "/usr/bin/lua",
           "-e",
           "NSGCONFD_HTTP_PORT="..tostring((httpconf.port or 80)),
           "/usr/bin/xavante_start"
        }
        C.fakeLastComTimestamp()
        C.restartDaemon(path, com, {noTimestamp = true, atExit = C.increasingSleep})
      else
        C.fakeLastComTimestamp()
        C.restartDaemon(path)
      end

      return true
    else
      si.EXTRA["http-apply-warn"] = true
      local _H = "servicesHttpApplyWarn"
      return nil,  s.localeString(_H, si and si.locale)
    end

  end
  return true
end

function getInetdApplyFunc(changeInetdConfFunc)
  return function(path, sch, conf, pcnf, first, ses)
    conf = s.addDefault(path, sch, conf, true) or {}
    local inetdConf = C.getOption({'services'}, 'inetdConf') or {}
    changeInetdConfFunc(inetdConf, conf)
    C.setOption({'services'}, 'inetdConf', inetdConf)
    local f, err = io.open("/etc/inetd.conf", 'w')
    if f then
      local inetdConfFile = ""
      for _, v in pairs(inetdConf) do
        inetdConfFile = inetdConfFile..v..'\n'
      end
      f:write(inetdConfFile)
      f:close()
      os.execute"/etc/init.d/inetd reload >/dev/null 2>&1"
      return true
    end
    return nil, err
  end
end

local function getHttpFilterRule(path)
   local rule = {
      protocol = "tcp",
      ["destination-port"] = "80",
      target = "DROP",
      _keep = true,
   }
   local filterPath = {'ip', 'filter', 'INPUT'}
   local filterInputList = C.getConfByPath(filterPath)
   if filterInputList then
      for k, v in ipairs(filterInputList) do
         if s.confCmp(rule, v) then
        return true, filterPath, k, rule
      end
    end
    return false, filterPath, #filterInputList + 1, rule
  else
    return false, filterPath, 1, rule
  end
end


local inetdInternalScheme = s.newScheme{
  udp = s.choiceType{_H = "servicesInternalUdp", _A={"enable","disable"}, _D = "disable"},
  tcp = s.choiceType{_H = "servicesInternalTcp", _A={"enable","disable"}, _D = "disable"},
}

servicesScheme = s.newScheme{
  _Z = function() return true end,
  _H = "services",

  ssh = s.newScheme{
    _H = "servicesSsh",
    _Z = restartSSHDaemon,

    enable = s.booleanType{_H = "servicesSshEnable"},
    ["debug-level"] = s.newScheme{
      _H = "servicesSshDebugLevel",
      _T = "number",
      _A = {0, 3},
      _V = s.integerVal,
      _D = 0,
    },
    port = s.tcpPortType{
      _H = "servicesSshPort",
      _D = 22,
    },
    ["log"] =  s.newScheme{
      _H = "servicesSshLog",
      _X = function(path)
        path = s.clonePath(path, -1)
        return C.getDaemonLog(path)
      end,
      _R = 'user',
    },
    keygen = s.newScheme{
      _H = "servicesSshKeygen",
      _T = "string",
      _EX = { subtype = 'areyousure'},
      _V = s.areyousureVal,
      _A = {"\nNew RSA snd DSA host keys will be generated.\n"..
      "It takes a lot of time.\n"},
      _X = function(path, sch, val, sesNum)
        local com
        local yes = val == 'yes'
        if yes then
          com = {'/sbin/keygen-ssh'}
        else
          com = {'echo', 'RSA snd DSA host keys NOT generated.'}
        end
        local myLog = ''
        C.restartDaemon(path, com,
          {sesNum = sesNum, noTimestamp = true, noRestart = true}
        ) 
      end,
      _XI = true,
    },
  },

  http = s.newScheme{
    _H = "servicesHttp",
    _Z = restartHttpDaemon,
    _N = function(sch, conf)
      if (C.getConfOrDefault({"services", "http","https"}) == "enable") then
        return {}
      else
        return { ["renew-cert"] = 1 }
      end
    end,
    enable = s.booleanType{_H = "servicesHttpEnable", _D = true, _DEP = 10},
    port = s.tcpPortType{
       _H = "servicesHttpPort",
       _D = 80,
    },
--[[
     home = s.stringType{
        _H = "servicesHttpHome",
        _D = "/usr/share/www/"
     },
     --  ]]
    ["log"] =  s.newScheme{
      _H = "servicesHttpLog",
      _X = function(path)
        path = s.clonePath(path, -1)
        return C.getDaemonLog(path)
      end,
      _R = 'user',
    },
    ["https"] = s.choiceType{
      _H = "servicesHttpHTTPS",
      _A = {
        {"enable", ""},
        {"disable", ""},
      },
      _D = "disable",
      _DEP = 10,
    },
    ["renew-cert"] = s.newScheme{
      _H = "servicesHttpRenewCert",
      _X = function(path,  sch, conf, sesNum)
        local si = C.getSessionInfo(sesNum)
        --     if si.rights=="admin" then
        C.log_warn(s.ser(si))
        if si.EXTRA["https-renew-cert-warn"] or si.mode ~= "html" then
          os.execute("rm -f /etc/stunnel/https_auto.cert; rm -f /etc/stunnel/https_auto.key")
          local _H = "servicesHttpRenewCertWarnAfter"
          return s.localeString(_H, si.locale)
        else
          si.EXTRA["https-renew-cert-warn"] = true
          local _H = "servicesHttpRenewCertWarnBefore"
          return s.localeString(_H, si.locale)
        end
     end,
   },
    ["deny-http"] = function(_, path)
                       if C.getConfByPath(s.clonePath(path, -1, 'https')) ~= 'enable' then
                          return nil
                       end
                       return not getHttpFilterRule(path) and
                       s.newScheme{
                       _H = "httpDenyHttp",
                       _X = function(path, sch, val, sesNum)
                               local si = sesNum and C.getSessionInfo(sesNum) or {}
                               local res, filterPath, num, rule = getHttpFilterRule(path)
                               if res then
                                  local _H = "httpMesRuleAllreadyExists_S_"
                                  return s.localeString(_H, si.locale):format(table.concat(filterPath, '.')..'.'..num)
                               else
                                  C.setConfigValue(s.clonePath(filterPath, nil, num), rule)
                                  local _H = "httpMesNewRuleAddTo_S_DontForgotApply"
                                  return s.localeString(_H, si.locale):format(table.concat(filterPath, '.')..'.'..num)
                               end
                            end,
                       _DEP = 3,
                    }
                 end,

    ["allow-http"] = function(_, path)
                        return getHttpFilterRule(path) and
                        s.newScheme{
                        _H = "httpAllowHttp",
                        _X = function(path,sch,val,sesNum)
                                local si = sesNum and C.getSessionInfo(sesNum) or {}
                                local res, filterPath, num, rule = getHttpFilterRule(path)
                                if res then
                                   C.removeConfigValue(s.clonePath(filterPath, nil, num))
                                   local _H = "httpMesRuleDeleted_S_"
                                   return s.localeString(_H, si.locale):format(table.concat(filterPath, '.')..'.'..num)
                                else
                                   local _H = "httpMesNoRuleToDeleteIn_S_"
                                   return s.localeString(_H, si.locale):format( table.concat(filterPath, '.'))
                                end
                             end,
                        _DEP = 3,
                     }
                  end,
 },

  telnet = s.newScheme{
    _H = "servicesTelnet",
    _Z = (function()
        local telnetKey = {}
        return getInetdApplyFunc(
          function(inetdConf, conf)
            inetdConf[telnetKey] = conf.enable
            and ("%d stream tcp nowait root /usr/sbin/telnetd telnetd"):
            format(conf.port) or nil
          end
        )
    end)(),

    enable = s.booleanType{_H = "servicesTelnetEnable", _D = true},
    port = s.tcpPortType{
      _H = "servicesTelnetPort",
      _D = 23,
    },
  },
  echo = inetdInternalScheme{
    _H = "servicesEcho",
    _Z = (function()
        local echoUdpKey = {}
        local echoTcpKey = {}
        return getInetdApplyFunc(
          function(inetdConf, conf)
            inetdConf[echoUdpKey] = conf.udp == 'enable'
            and "echo dgram udp wait root internal" or nil
            inetdConf[echoTcpKey] = conf.tcp == 'enable'
            and "echo stream tcp nowait root internal" or nil
          end
        )
    end)(),
  },
  discard = inetdInternalScheme{
    _H = "servicesDiscard",
    _Z = (function()
        local discardUdpKey = {}
        local discardTcpKey = {}
        return getInetdApplyFunc(
          function(inetdConf, conf)
            inetdConf[discardUdpKey] = conf.udp == 'enable'
            and "discard dgram udp wait root internal" or nil
            inetdConf[discardTcpKey] = conf.tcp == 'enable'
            and "discard stream tcp nowait root internal" or nil
          end
        )
    end)(),
  },
  chargen = inetdInternalScheme{
    _H = "servicesChargen",
    _Z = (function()
        local chargenUdpKey = {}
        local chargenTcpKey = {}
        return getInetdApplyFunc(
          function(inetdConf, conf)
            inetdConf[chargenUdpKey] = conf.udp == 'enable'
            and "chargen dgram udp wait root internal" or nil
            inetdConf[chargenTcpKey] = conf.tcp == 'enable'
            and "chargen stream tcp nowait root internal" or nil
          end
        )
    end)(),
  },
  daytime = inetdInternalScheme{
    _H = "servicesDaytime",
    _Z = (function()
        local daytimeUdpKey = {}
        local daytimeTcpKey = {}
        return getInetdApplyFunc(
          function(inetdConf, conf)
            inetdConf[daytimeUdpKey] = conf.udp == 'enable'
            and "daytime dgram udp wait root internal" or nil
            inetdConf[daytimeTcpKey] = conf.tcp == 'enable'
            and "daytime stream tcp nowait root internal" or nil
          end
        )
    end)(),
  },

  dhcp = dhcp.DhcpScheme,
  ["dns"] = dhcp.DnsServerScheme,
  ["dynamic-dns"] = dyndns.DynDnsServerScheme,
  upnp = upnp.UPnPServerScheme,
  netping = netping.NetPingScheme,
  daemons = daemons.Daemons,
  snmp = snmp.snmpScheme,
  ['event-handler'] = event.eventScheme,
  ['event-server'] = event.eventServerScheme,
  ['ixia-endpoint'] = ixia.endpointScheme,
  tcpproxy = tp.tcpproxyScheme,
  netflow = netflow.NetflowScheme,
}
