local nsg = require'nsg' nsg.meta()
local ox = require'posix'
local curses = require'curses'
local signal = require'signal'
local M = {}
local first = true

local hasColors

local exec
local reDraw

local w  -- main window
local witem
local whelp
local witemHeight
local witemWidth
local whelpWidth
local whelpHeight

local menu
local row, col, vert, horiz
local curpos = 1
local parentCurpos = {}
local startItem = 1

local globalDescr
local timeoutHook


local exitCurses
local ses, path
local debugMode = false
local min = math.min

local termfunc
local aboutMes
local startDescr

local function printw(...)
  local t = {...}
  w:move(0,0)
  for i = 1, select('#', ...) do
    w:addstr(tostring(t[i]))
    w:addstr'  '
  end
  w:addstr'\n'
  w:getch()
end

local function utf8len(ustring)
  local _, count = string.gsub(ustring, "[^\128-\193]", "")
  return count
end

local function utf8index(s, k)
  local b, n = 0, 0
  local over = #s + 1
  for i = 1, k do
    b, n = s:find("([%z\1-\127\194-\244][\128-\191]*)", n + 1)
    if not n then break end
  end
  return b or over, n or over
end

local function utf8sub(str, b, e)
  local x, y
  if e then
    x, y = utf8index(str, e)
  end
  return str:sub(utf8index(str, b), y)
end

local function getc(self)
  local c, char = self:getch()
  if c and c < 256 then char = string.char(c) end
  return c, char
end

----------


local PAIR_BLACK = 1
local PAIR_RED = 2
local PAIR_GREEN = 3
local PAIR_YELLOW = 4
local PAIR_BLUE = 5
local PAIR_MAGENTA = 6
local PAIR_CYAN = 7
local PAIR_WHITE = 8

local colorschemeDefault = {
    background = "black",
    normal = "white",
    folder = "cyanB",
    command = "green",
    commandWithPar = "green",
    interactiveCommand = "magenta",
    interactiveCommandWithPar = "magenta",
    parameter = "blackB",
    emptyParameter = "black",
}

local c2c
local function set_c2c()
  c2c = {
    black = curses.COLOR_BLACK,
    red = curses.COLOR_RED,
    green = curses.COLOR_GREEN,
    yellow = curses.COLOR_YELLOW,
    blue = curses.COLOR_BLUE,
    magenta = curses.COLOR_MAGENTA,
    cyan = curses.COLOR_CYAN,
    white = curses.COLOR_WHITE,
  }
end

local c2p = {
  black = curses.color_pair(PAIR_BLACK),
  red = curses.color_pair(PAIR_RED),
  green = curses.color_pair(PAIR_GREEN),
  yellow = curses.color_pair(PAIR_YELLOW),
  blue = curses.color_pair(PAIR_BLUE),
  magenta = curses.color_pair(PAIR_MAGENTA),
  cyan = curses.color_pair(PAIR_CYAN),
  white = curses.color_pair(PAIR_WHITE),
}

local colorscheme = {
  {
  },
  {
    normal = "black",
  },
}
local colornum = 1

local colorAttr = {}

local function getcsattr(attr)
  local color, bold = attr:match"(.-)(B?)$"
  return c2p[color] + (bold == 'B' and curses.A_BOLD or 0)
end

local function initColors()
  if hasColors then
    curses.use_default_colors()
    local fgDefault, bgDefault = curses.pair_content(0)
    local cs = colorscheme[colornum] or{}
    setmetatable(cs, {__index = colorschemeDefault})
    --local bgDefault = c2c[cs.background]
    curses.init_pair(PAIR_BLACK, -1, -1)
    curses.init_pair(PAIR_RED, curses.COLOR_RED, -1)
    curses.init_pair(PAIR_GREEN, curses.COLOR_GREEN, -1)
    curses.init_pair(PAIR_YELLOW , curses.COLOR_YELLOW, -1)
    curses.init_pair(PAIR_BLUE, curses.COLOR_BLUE, -1)
    curses.init_pair(PAIR_MAGENTA, curses.COLOR_MAGENTA, -1)
    curses.init_pair(PAIR_CYAN, curses.COLOR_CYAN, -1)
    curses.init_pair(PAIR_WHITE, -1, -1)

    colorAttr.b = getcsattr(cs.folder)
    colorAttr.c = getcsattr(cs.command)
    colorAttr.k = getcsattr(cs.commandWithPar)
    colorAttr.i = getcsattr(cs.interactiveCommand)
    colorAttr.j = getcsattr(cs.interactiveCommandWithPar)
    colorAttr.p = getcsattr(cs.parameter)
    colorAttr.q = getcsattr(cs.emptyParameter)
    colorAttr.n = getcsattr(cs.normal)
  end
  setmetatable(colorAttr, {__index = function() return curses.color_pair(PAIR_RED) end})
end


local fullHelp = [[
ARROW RIGHT, ENTER - Next menu node
ARROW LEFT         - Previous menu node
TAB                - Windows toggle
^R                 - Refresh screen

a   - About
e   - Edit configuration in editor mode
h   - Show help
l   - Set locale(e.g 'en', 'ru')
m   - Show manual for this parameter
o   - Set output mode for command _print, _show, ...
p   - Print current node configuration(without defaults)
q   - Quit 
r   - Remove this node
s   - Show current node configuration(with defaults)
t   - Open Linux terminal
u   - Show system usage
w   - Save configuration

Press TAB to return to the main window.
]]

local mainHelp = "TAB-toggle h-help m-man l-locale p-print r-remove w-write q-exit"
local readlineHelp = "ENTER-complete input"
local helpStr = mainHelp

local function drawHelp()
  local str = helpStr..(' '):rep(col - #helpStr)
  w:attron(curses.A_REVERSE)
  w:mvaddstr(row - 1, 0, str)
  --w:mvaddch(row - 1, 32, curses.ACS_RARROW)
  w:attroff(curses.A_REVERSE)
end

local function addDescr(str)
  w:move(horiz + 1, 0)
  w:clrtobot()
  w:mvaddstr(horiz + 1, 0, str)
  drawHelp()
  w:refresh()
end

local function drawCursor()
  --globalDescr = nil
  local cursorPos = curpos - startItem
  if cursorPos >= 0 then
    local menupos = cursorPos + startItem
    witem:attron(curses.A_REVERSE + colorAttr[menu[menupos].itype])
    witem:mvaddstr(cursorPos, 0, utf8sub(menu[menupos].item, 1, witemWidth))
    witem:attroff(curses.A_REVERSE + colorAttr[menu[menupos].itype])
    witem:refresh()
    --  w:mvaddch(menupos, 0, curses.ACS_RARROW )
    local helpstr = menu[menupos].help
    local helpstrlen = utf8len(helpstr)
    if helpstrlen > whelpWidth then
      w:timeout(2000)
      local state = 1
      whelp:mvaddstr(cursorPos, 0, utf8sub(helpstr, 1, whelpWidth))
      whelp:refresh()
      timeoutHook = function()
	state = state + 1
	whelp:mvaddstr(cursorPos, 0, utf8sub(helpstr, state, whelpWidth + state - 1))
	whelp:refresh()
	if whelpWidth + state - 1 >= helpstrlen then
	  state = 0
	  w:timeout(2000)
	elseif state == 1 then
	  w:timeout(2000)
	else
	  w:timeout(200)
	end
      end
    end

    addDescr''
    local str, _ = globalDescr or menu[menupos].descr
    if str then
      addDescr(str)
    else
      local typ = menu[menupos].itype
      if typ == 'p' or typ == 'q' then
        if not str then
          local item = menu[menupos].item
	  item = tonumber(item) or item
          path[#path+1] = item
          _, str = ses:setValueSmart(path)
          if str then
	    if str:find"expected variants:" or str:find"expected boolean" then
	      menu[menupos].choice = true
	    end
            str = str:match"\n(curr = .*)$" or str
            menu[menupos].descr = str:match'^%s*(.-)%s*$'
            addDescr(str)
	    if typ == 'p' then
	      str = str:match"curr =%s*([^%s][^\n]-)%s*\n"
	      if str then
		menu[menupos].parmval = str
	      end
	    end
          end
          path[#path] = nil
        end
      end
    end
  end
end

local function clearCursor()
  local cursorPos = curpos - startItem
  if cursorPos >= 0 then
    local menupos = cursorPos + startItem
    --  w:mvaddch(menupos, 0, curses.ACS_VLINE)
    witem:attron(colorAttr[menu[menupos].itype])
    witem:mvaddstr(cursorPos, 0, utf8sub(menu[menupos].item, 1, witemWidth))
    witem:attroff(colorAttr[menu[menupos].itype])
    witem:refresh()
    whelp:mvaddstr(cursorPos, 0, utf8sub(menu[menupos].help, 1, whelpWidth))
    whelp:refresh()
  end
end

local function setStartItem()
  if curpos > #menu then
    curpos = #menu
  end
  if curpos < startItem then
    startItem = curpos
    return true
  elseif curpos >= witemHeight + startItem then
    startItem = curpos - witemHeight + 1
    return true
  else
    return false
  end

end

local function setCursor()
  globalDescr = nil
  if setStartItem() then
    reDraw()
  else
    drawCursor()
  end
end

local function getmaxyx()
  return w:getmaxyx()
end
local function clearScreen()

  if witem then witem:close() end
  if whelp then whelp:close() end
  if w then w:close() end
  w = curses.stdscr()
  w:keypad(true)
  w:clear()
  row, col = getmaxyx()
  w:clear()
end

function reDraw()
  clearScreen()

  if row < 7 or col < 9 then
    local warn = "Too small screen"
    if row * col < #warn then
      warn = ('*'):rep(row * col)
    end
    w:mvaddstr(0, 0, warn)
    return
  end

  vert = math.floor(col * 0.33)
  horiz = math.floor(row * 0.67)
  w:mvaddch(0, 0, curses.ACS_ULCORNER)
  w:mvaddch(0, col - 1, curses.ACS_URCORNER)
  w:mvaddch(horiz, 0, curses.ACS_LLCORNER)
  w:mvaddch(horiz, col - 1, curses.ACS_LRCORNER)
  w:mvvline(1, 0, 0, horiz - 1)
  w:mvvline(1, vert, 0, horiz - 1)
  w:mvvline(1, col - 1, 0, horiz - 1)
  w:mvhline(0, 1, 0, col - 2)
  w:mvhline(horiz, 1, 0, col - 2)
  --  w:mvaddch(0, vert, curses.ACS_TTEE)
  --  w:mvaddch(horiz, vert, curses.ACS_BTEE)

  local prompt = ox.uname'%n'..':'..path:concat('.')
  if #prompt > col - 4 then
    prompt = "..."..prompt:sub(-(col - 7), -1)
  end
  w:attron(curses.A_BOLD)
  w:mvaddstr(0, 2, prompt)
  w:attroff(curses.A_BOLD)

  witemHeight = horiz - 1
  witemWidth = vert - 1
  whelpHeight = horiz - 1
  whelpWidth = col - vert - 2
  witem = curses.newwin(witemHeight, witemWidth, 1, 1)
  whelp = curses.newwin(whelpHeight, whelpWidth, 1, vert + 1)


  setStartItem()
  if startItem > 1 then
    w:attron(curses.A_BOLD)
    w:mvaddch(1, 0, curses.ACS_UARROW)
    w:attroff(curses.A_BOLD)
  end
  if startItem + witemHeight <= #menu  then
    w:attron(curses.A_BOLD)
    w:mvaddch(horiz - 1, 0, curses.ACS_DARROW)
    w:attroff(curses.A_BOLD)
  end
  for i = 0, min(witemHeight - 1, #menu - startItem) do  
    local menupos = i + startItem
    local attr = colorAttr[menu[menupos].itype]
    witem:attron(attr)
    witem:mvaddstr(i, 0, utf8sub(menu[menupos].item, 1, witemWidth))
    witem:attroff(attr)
    whelp:mvaddstr(i, 0, utf8sub(menu[menupos].help, 1, whelpWidth))
  end
  drawCursor()
  witem:overlay(w)
  whelp:overlay(w)
  curses.curs_set(0)


end

local function initMainWindow()
  if S.leaf() then
    S.popPath()
  end
  local com, str = ses:setValueSmart(path)
  menu = {{itype = 'b', item = '..', help = ""}}
  for itype, item, help in str:gmatch'(%a)([^%s]+)%s+([^\n]*)\n' do
    menu[#menu+1] = {itype = itype, item = item, help = help}
  end
  globalDescr = nil
  reDraw()
  if #menu == 1 then
    globalDescr = str:match'^%s*(.-)%s*$'
    addDescr(globalDescr)
  end
end

local descrCur
local descrLines
local descrString

local function reDrawDescr()
  w:clear()
  w:mvaddstr(0, 0, descrString:sub(descrLines[descrCur]))
end

local function drawDescr(tail)
  descrCur = 1
  descrLines = {}
  local a, b = 0, 0
  repeat
    local prev = b + 1
    descrLines[#descrLines+1] = prev
    a, b = descrString:find('\n', prev)
    if a then
      while prev + col <= b do
	prev = prev + col
	descrLines[#descrLines+1] = prev
      end
    end
  until not a
  clearScreen()
  if tail then
    descrCur = #descrLines - row + 2
    if descrCur < 1 then
      descrCur = 1
    end
  end
  reDrawDescr()
end

local function writeConf()
  local cnf, err = ses:getConf(path, true)
  if not cnf then return err end

  local lcnf = ses:getLocalConf(path, true)
  if not lcnf then return err end

  lcnf = (not lcnf:match"^%s*$") 
  and "\nLocal configuration:\n"..lcnf or ''
  return cnf..lcnf
end

local function readline(w, y, x, sz, prompt, as)
  w:attron(curses.A_BOLD)
  curses.curs_set(1)

  prompt = prompt or ''
  local pl = utf8len(prompt)
  local size = sz - pl
  w:move(y, x)
  w:addstr(prompt)


  local b = x + pl
  local e = b + size

  local as = as or ""
  local pos = utf8len(as)
  local offs = 0

  local wasReverse
  if #as > 0 then
    w:attron(curses.A_REVERSE)
    wasReverse = true
  end

  local n, ch, resize
  while true do
    if pos < offs then
      offs = pos
    elseif pos >= offs + size then
      offs = pos - size + 1
    end

    w:mvaddstr(y, b, utf8sub(as, 1 + offs, size + offs))
    w:clrtoeol()
    if offs > 0 then
      w:mvaddch(y, b, curses.ACS_LARROW)
    end
    if utf8len(as) - offs > size then
      w:mvaddch(y, e - 1, curses.ACS_RARROW)
    end
    w:move(y, b + pos - offs)
    
    repeat n, ch = getc(w) until n
    if n >= 194 and n <= 244 then -- UTF-8
      local _, c = getc(w)
      ch = ch..c
    end

    if n == 10 then
      break
    elseif n == 27 then
      as = false
      break
    elseif n == 263 then -- bs
      if pos > 0 then
        as = utf8sub(as, 1, pos - 1)..utf8sub(as, pos + 1)
        pos = pos - 1
        if offs > 0 then
          offs = offs - 1
        end
      end
    elseif n == 330 then -- delete
      as = utf8sub(as, 1, pos)..utf8sub(as, pos + 2)
    elseif n == 260 then --  <-
      if pos > 0 then
        pos = pos - 1
      elseif as == '' then
	as = false
	break
      end
    elseif n == 261 then --  ->
      if pos < utf8len(as) then
        pos = pos + 1
      end
    elseif n == 410 then
      resize = true
      break
    elseif ch and (#ch == 2 or ch:find"[^%c]") then
      if wasReverse then
	as, pos, offs = "", 0, 0
      end
      as = utf8sub(as, 1, pos)..ch..utf8sub(as, pos + 1)
      pos = pos + 1
    end
    if wasReverse then
      wasReverse = false
      w:attroff(curses.A_REVERSE)
    end
  end
  w:attroff(curses.A_REVERSE)

  curses.curs_set(0)
  w:attroff(curses.A_BOLD)

  return as, resize
end

local function empty() end

local function exitCursesFunc(char)
  w:clear()
  w:refresh()
  curses.endwin()
  exitCurses = true
end

local function notImplemented(char)
  globalDescr = "This feature is not implemented yet - "..char
  addDescr(globalDescr)
  --S.popPath()
end

local function readlineMain(prompt, initial)
  helpStr = readlineHelp
  addDescr''
  helpStr = mainHelp
  return readline(w, horiz + 1, 0, col, prompt..' = ', initial)
end


local function drawChoice(choice)
  addDescr''
  local c, r = choice.curr[1], choice.curr[2]
  if c < choice.startcol then choice.startcol = c end
  local offs = choice[choice.startcol].x - 1
  choice.endcol = #choice
  for c = choice.startcol, #choice do
    local column = choice[c]
    if column.x - offs + column.size > col - 1 and c > choice.startcol then 
      choice.endcol = c - 1
      break 
    end
    for k, v in ipairs(column) do
      w:mvaddstr(horiz + k, column.x - offs, v:sub(1, col - 1 - column.x + offs))
    end
  end
  if c > choice.endcol then 
    choice.startcol = choice.startcol + 1 
    return drawChoice(choice)
  end
  w:attron(curses.A_REVERSE)
  w:mvaddstr(horiz + r, choice[c].x - offs, choice[c][r]:sub(1, col - 1 - choice[c].x + offs))
  w:attroff(curses.A_REVERSE)
  if choice.startcol > 1 then
    for i = 1, #choice[1] do
      w:mvaddch(horiz + i, 0, curses.ACS_LARROW)
    end
  end
  if choice.endcol < #choice then
    for i = 1, #choice[1] do
      w:mvaddch(horiz + i, col - 1, curses.ACS_RARROW)
    end
  end
end

local function drawChoiceCursor(choice, newCol, newRow)
  local cl, row = choice.curr[1], choice.curr[2]
  choice.curr = {newCol, newRow}
  if not choice.endcol or newCol < choice.startcol or newCol > choice.endcol then
    return drawChoice(choice)
  end
  local offs = choice[choice.startcol].x - 1
  local bold = choice[cl].bold == row
  if bold then w:attron(curses.A_BOLD) end
  w:mvaddstr(horiz + row, choice[cl].x - offs, choice[cl][row]:sub(1, col - 1 - choice[cl].x + offs))
  if bold then w:attroff(curses.A_BOLD) end
  w:attron(curses.A_REVERSE)
  w:mvaddstr(horiz + newRow, choice[newCol].x - offs, choice[newCol][newRow]:sub(1, col - 1 - choice[newCol].x + offs))
  w:attroff(curses.A_REVERSE)
end

local function readChoice(path, curr)
  local _, str = ses:setValueSmart(path)
  if str then
    local valstr = str:match"expected variants:(.-)\ncurr"
    if not valstr then
      if str:find"expected boolean" then
	valstr = "true\nfalse\n"
      else
	return
      end
    end

    local choice = {{x = 0}, startcol = 1, curr = {1,1}}
    local high = row - horiz - 2
    local column, n = 1, 1
    local offset, max = 0, 0
    for item in valstr:gmatch"%s*([^%s][^\n]-)%s*\n" do
      if item == curr then 
	choice.curr = {column, n}
	choice[column].bold = n
      end
      if item:sub(1,1) == '"' then
	item = item:sub(2,-2):gsub("\\\"", '"')
      end
      choice[column][n] = item
      n = (n + 1)%(high + 1)
      if #item > max then max = #item end
      if n == 0 then
	n = 1
	choice[column].size = max
	column = column + 1
	offset = offset + max + 1
	choice[column] = {x = offset}
	max = 0
      end
    end
    choice[column].size = max


    drawChoice(choice)
    local col, row = choice.curr[1], choice.curr[2]
    while true do
      repeat n, ch = getc(w) until n
      if n == 259 then      -- up
	row = row -1 
	if row < 1 then
	  row = 1
	end
      elseif n == 258 then  -- down
	row = row + 1
	if row > #choice[col] then
	  row = #choice[col]
	end
      elseif n == 261 then  -- right
	col = col + 1
	if col > #choice then
	  col = #choice
	end
	if row > #choice[col] then
	  row = #choice[col]
	end
      elseif n == 260 then  -- left
	col = col - 1
	if col < 1 then
	  col = 1
	end
      elseif n == 10 then
	break
      elseif n == 27 then
	return false
      end
      drawChoiceCursor(choice, col, row)
    end
    return choice[choice.curr[1]][choice.curr[2]]
  end
end

local execDescr
local execMain = {
  [-1] = empty,         -- ERR or TIMEOUT

  [4] = function(char)  -- ^D - toggle debug mode
    debugMode = not debugMode
    w:clear()
    if debugMode then
      w:mvaddstr(0, 0, "***** DEBUG MODE *****   (press ^D to exit debug mode)\n")
    end
  end,

  [9] = function(char)  -- TAB
    descrString = globalDescr or menu[curpos].descr or ''
    if descrString and descrString ~= '' then
      drawDescr()
      exec = execDescr
    end
  end,

  [353] = function(char)  -- S-TAB
    descrString = globalDescr or menu[curpos].descr or ''
    if descrString and descrString ~= '' then
      drawDescr(true)
      exec = execDescr
    end
  end,


  [10] = function(char)  -- ENTER
    local item = menu[curpos].item
    if item == ".." then
      curpos = parentCurpos[#path] or 1
      S.popPath()
      initMainWindow()
      return
    end
    item = tonumber(item) or item
    path:insert(item)

    local it = menu[curpos].itype
    if it == 'b' then
      S.leaf(false)
      parentCurpos[#path] = curpos
      curpos = 1
      initMainWindow()
    elseif it == 'c' then
      menu[curpos].descr = nil
      local _, str = ses:setValueSmart(path)
      if str then
        str = str:match'^%s*(.-)%s*$'
	str = str:gsub('\r\n', '\n')
        menu[curpos].descr = str
        addDescr(str)
      end
      S.popPath()

    elseif it == 'p' or it == 'q' or it == 'k' then
      local res, str = nil, ""
      local val, resize 
      if menu[curpos].choice then
	val, resize = readChoice(path, menu[curpos].parmval)
      else
	val, resize = readlineMain(item, menu[curpos].parmval)
      end
      if val then
	val = S.eval(val)
	res, str = ses:setValueSmart(path, val)
      end
      S.popPath()
      menu[curpos].descr = nil
      if resize then reDraw() end
      if str then
	local gd = str:match'^%s*(.-)%s*$'
	if it == 'k' then
	  menu[curpos].descr = gd
	elseif gd:sub(1,1) == '=' then
	  initMainWindow() -- menu can changed
	end
	globalDescr = gd
	addDescr(globalDescr)
      end

    elseif it == 'i' or it == 'j' then
      local com, err, arg, ttyPath
      if it == 'i' then
	local s = path[#path]
	S.popPath()
	com, err, arg, ttyPath = ses:setValueSmart(path, s)
      else
	local val, resize = readlineMain(item, menu[curpos].parmval)
	if val then
	  val = S.eval(val)
	  com, err, arg, ttyPath = ses:setValueSmart(path, val)
	  if err then
	    globalDescr = err:match'^%s*(.-)%s*$'
	    addDescr(globalDescr)
	  end
	  S.popPath()
	end
      end
      if ttyPath then 
	termfunc = function()
	  io.write('Interactive mode.\n'..
	           'Type ^C^K to kill running process.\n'
	          )
	  io.flush()
	  S.imode(ttyPath)
	end
	exitCursesFunc()
      end

    -- _new and _insert
    elseif it == 'n' then
      local com = path[#path]
      S.popPath()

      if com == "_apply" then
	local val, resize = readlineMain("Apply config for node "..
	                                 table.concat(path, '.').."?[YES/no] ")
	val = val or 'no'
	val = val:match"^%s*([^%s]*)%s*$"
	if val == ("yes"):sub(1, #val) then
	  globalDescr = "Start apply."
	  addDescr(globalDescr)
	  local ret, err = ses:applyConf(path)
	  globalDescr = ret or err
	  addDescr(globalDescr)
	else
	  globalDescr = "Don't apply."
	  addDescr(globalDescr)
	end
      else -- _new or _insert
	local method = com == '_new' and "newKey" or "insertKey"

	local val, resize = readlineMain(item)
	local ret, err = nil, ""
	if val then
	  val = S.eval(val)
	  ret, err, val = ses[method](ses, path, val)
	end
	if ret then
	  curpos = curpos + 1
	  initMainWindow()
	else
	  globalDescr = err:match'^%s*(.-)%s*$'
	  addDescr(globalDescr)
	end
      end

    else
      S.popPath()
    end
  end,

  [18] = function(char)     -- ^R - redraw
    reDraw()
  end,

  [97] = function()         -- a - about
    globalDescr = aboutMes
    addDescr(globalDescr)
  end,

  [101] = function()        -- e - edit
    local cnf, err = ses:getConf(path, false)
    if not cnf then
      globalDescr = cnf:match'^%s*(.-)%s*$'
      addDescr(globalDescr)
      return
    end

    termfunc = function()
      local filename = os.tmpname()
      local f = io.open(filename, 'w')
      f:write(tostring(cnf))
      f:close()

      os.execute("nano -L "..filename)

      local f = io.open(filename)
      cnf = f:read'*a'
      f:close()
      os.remove(filename)

      local cnf, err = ses:paste(path, cnf..'\n')
      startDescr = cnf and 'OK' or err
    end
    exitCursesFunc()
  end,

  [104] = function() -- h - help
    globalDescr = fullHelp
    addDescr(globalDescr)
    descrString = globalDescr -- goto full descr window
    drawDescr()
    exec = execDescr
  end,

  [108] = function()   -- l - locale
    local val, resize = readlineMain("locale")
    globalDescr = ''
    if val then
      local ret, err = ses:setLocale(val)
      initMainWindow()
      globalDescr = err or 'OK'
    end
    addDescr(globalDescr)
  end,

  [109] = function(char)  -- m - manual
    local item = menu[curpos].item
    item = tonumber(item) or item
    path:insert(item)
    local hlp = ses:getManual(path) or "No manual."
    globalDescr = hlp:match'^%s*(.-)%s*$'
    addDescr(globalDescr)
    S.popPath()
  end,

  [111] = function()   -- o - output mode
    local val, resize = readlineMain("output-mode")
    globalDescr = ''
    if val then
      if val == '' then
	val = '*'
      end

      local ret, err
      if val == ('raw'):sub(1, #val) then
	val = 'raw'
      elseif val == ('user-friendly'):sub(1, #val) then
	val = 'user-friendly'
      else
	err = 'expected "raw" or "user-friendly"'
      end
      if not err then
	ret, err = ses:setOutputMode(val)
      end
      globalDescr = err or 'OK'
    end
    addDescr(globalDescr)
  end,

  [112] = function(char)  -- p print
    local cnf, err = ses:getConf(path, false)
    globalDescr = (cnf or err):match'^%s*(.-)%s*$'
    addDescr(globalDescr)
  end,

  [113] = function() -- q - exit curses mode
    if ses:isConfigChanged() then
      local val, resize = readlineMain("Configuration not saved. Save it?[yes/no/cancel] ")
      val = val or ''
      if val == "yes" then
	local res, err = ses:saveConf()
      elseif val == ("cancel"):sub(1, #val) then
	return
      elseif val ~= "no" then
	globalDescr = "To exit press 'q' and answer 'yes' or 'no'"
	addDescr(globalDescr)
	return
      end
    end
    exitCursesFunc()
  end,

  [114] = function()   -- r - remove
    local item = menu[curpos].item
    item = tonumber(item) or item
    path:insert(item)
    local val, resize = readlineMain("Remove/Clear node "..
                                     table.concat(path, '.').."?[yes/no] ")
    val = val or ''
    val = val:match"^%s*([^%s]*)%s*$"
    if #val > 0 and val == ("yes"):sub(1, #val) then
      local ret, err = ses:removeKey(path)
      S.popPath()
      initMainWindow()
      globalDescr = ret and "OK" or err
      addDescr(globalDescr)
    else
      globalDescr = "Don't remove."
      addDescr(globalDescr)
      S.popPath()
    end
  end,

  [115] = function(char)  -- s - show
    globalDescr = writeConf():match'^%s*(.-)%s*$'
    addDescr(globalDescr)
  end,

  [116] = function()   -- t - term
    local perm, rights = ses:getPerm()
    if perm and rights == "admin" then
      termfunc = function()
	io.write('Type "exit" to return to nsg shell\n')
	io.flush()
	os.execute'sh --login'
      end
      exitCursesFunc()
    else
      globalDescr = not perm and rights or 
                    "You have no rights to open Linux terminal"
      addDescr(globalDescr)
    end
  end,

  [117] = function(char)  -- u - system usage
    local usg = ses:getUsage() or "No info."
    globalDescr = usg:match'^%s*(.-)%s*$'
    addDescr(globalDescr)
  end,

  [118] = notImplemented,   -- v 
  [119] = function()   -- w - write
    local ret, err = ses:saveConf()
    globalDescr = ret or err
    addDescr(globalDescr)
  end,


  [258] = function(char)  -- arrow down
    if curpos < #menu then
      clearCursor()
      curpos = curpos + 1
      setCursor()
    else
      curses.beep()
    end
  end,

  [259] = function(char)  -- arrow up
    if curpos > 1 then
      clearCursor()
      curpos = curpos - 1
      setCursor()
    else
      curses.beep()
    end
  end,

  [260] = function(char)  -- arrow left
    curpos = parentCurpos[#path] or 1
    S.popPath()
    initMainWindow()
  end,

  [338] = function(char)  -- page down
    if curpos < #menu then
      clearCursor()
      curpos = curpos + witemHeight
      if curpos > #menu then curpos = #menu end
      setCursor()
    else
      curses.beep()
    end
  end,

  [339] = function(char)  -- page up
    if curpos > 1 then
      clearCursor()
      curpos = curpos - witemHeight
      if curpos < 1 then curpos = 1 end
      setCursor()
    else
      curses.beep()
    end
  end,

}
execMain[410] = execMain[18]
execMain[261] = execMain[10]
execMain[17]  = execMain[113] -- ^Q


execDescr = {
  [9] = function(char)  -- TAB - switch to main window
    reDraw()
    exec = execMain
  end,

  [17] = exitCursesFunc, -- ^Q - exit curses mode

  [18] = function(char)  -- ^R - redraw
    reDrawDescr()
  end,

  [258] = function(char)  -- arrow down
    if descrCur < #descrLines then
      descrCur = descrCur + 1
      reDrawDescr()
    else
      curses.beep()
    end
  end,

  [259] = function(char)  -- arrow up
    if descrCur > 1 then
      descrCur = descrCur - 1
      reDrawDescr()
    else
      curses.beep()
    end
  end,

  [338] = function(char)  -- page down
    if descrCur < #descrLines then
      descrCur = descrCur + 20
      if descrCur > #descrLines then descrCur = #descrLines end
      reDrawDescr()
    else
      curses.beep()
    end
  end,

  [339] = function(char)  -- page up
    if descrCur > 1 then
      descrCur = descrCur - 20
      if descrCur < 1 then descrCur = 1 end
      reDrawDescr()
    else
      curses.beep()
    end
  end,
}
execDescr[410] = execDescr[18]
execDescr[27] = execDescr[9]



local outRowNumber = 0
local execmt = {
  __index = function(t, c)
    return function(char)
      curses.beep()
      if debugMode then
        w:mvaddstr(outRowNumber, 0, "*** c = %s, char = '%s'\n"%
          {tostring(c), tostring(char)})
        outRowNumber = (outRowNumber + 1) % 24
      end
    end
  end,
}
setmetatable(execMain, execmt)
setmetatable(execDescr, execmt)

local function timeoutCallbackCurses(ses, state)
  w:timeout(0)
  repeat
    local c = getc(w)
  until not c
  w:timeout(-1)
  globalDescr = "Operation too long. Waiting "..('#'):rep(state)
  addDescr(globalDescr)
end

function M.nsgshcurses(rl)
  if first then
    ox.setenv("TERMINFO", "/usr/share/terminfo/", true)
    ox.setenv("TERM", "linux", true)
    ox.setenv("ESCDELAY", "1", true)
    first = nil
    os.setlocale('ru_RU.utf-8')
    curses.initscr()
    curses.raw()
    curses.echo(false)
    curses.start_color()
    hasColors = curses.has_colors()
    set_c2c()
    w = curses.stdscr()
    w:keypad(true)
    w:clear()
  end
  --[[
  if rl then -- readline was called
    row, col = w:getmaxyx()
    getmaxyx = function() return row, col end
    local function sigwinchhandler()
      local f = io.popen"stty -a"
      if f then
        local s = f:read'*a'
        newrow, newcol = s:match"rows%s+(%d+);%s+columns%s+(%d+);"
        row = tonumber(newrow) or row
        col = tonumber(newcol) or col
        reDraw()
      end
    end
    signal.signal('WINCH', sigwinchhandler)
  end
  --]]

  local version = S.ses:getVersion() or "UNKNOWN"
  aboutMes = [[
NSG Linux 2.0 build %s

%s
]]%{version, S.initialError}
  startDescr = aboutMes

  while true do
    ses, path = S.ses, S.path
    exitCurses = nil
    exec = execMain
    ses:setCurses(true)
    initColors()
    initMainWindow()
    if startDescr then
      globalDescr = startDescr
      addDescr(globalDescr)
      startDescr = nil
    end

    local timeoutCallbackOrig = ses.timeoutCallback
    ses.timeoutCallback = timeoutCallbackCurses

    repeat
      local c, char = getc(w)
      if timeoutHook then
	if c then
	  w:timeout(-1)
	  timeoutHook = nil
	else
	  timeoutHook()
	end
      end
      c = c or -1
      exec[c](char)
    until exitCurses
    ses:setCurses(false)
    ses.timeoutCallback = timeoutCallbackOrig
    os.execute("stty sane")
    if termfunc then
      termfunc()
      termfunc = nil
    else
      break
    end
  end
end

function M.close()
  w:close()
  curses.endwin()
  first = true
  ses:setCurses(false)
end


return M
